import pytest
import torch
from torch import nn
from torch.utils.data import DataLoader, TensorDataset

from tests._dummies.model import dummy_model
from torch_uncertainty.models import SWA, SWAG


class TestSWA:
    """Testing the SWA class."""

    def test_training(self):
        dl = DataLoader(TensorDataset(torch.randn(1, 1)), batch_size=1)
        swa = SWA(dummy_model(1, 10), cycle_start=1, cycle_length=1)
        swa.eval()
        swa(torch.randn(1, 1))

        swa.train()
        swa(torch.randn(1, 1))
        swa.update_wrapper(0)
        swa.bn_update(dl, "cpu")

        swa.update_wrapper(1)
        swa.bn_update(dl, "cpu")

        swa.eval()
        swa(torch.randn(1, 1))

    def test_failures(self):
        with pytest.raises(ValueError, match="`cycle_start` must be non-negative."):
            SWA(nn.Module(), cycle_start=-1, cycle_length=1)
        with pytest.raises(ValueError, match="`cycle_length` must be strictly positive."):
            SWA(nn.Module(), cycle_start=1, cycle_length=0)


class TestSWAG:
    """Testing the SWAG class."""

    def test_training(self):
        dl = DataLoader(TensorDataset(torch.randn(1, 1)), batch_size=1)
        swag = SWAG(
            dummy_model(1, 10),
            cycle_start=1,
            cycle_length=1,
            max_num_models=3,
            num_estimators=2,
        )
        assert swag.num_avgd_models == 0
        swag.eval()
        swag(torch.randn(1, 1))

        swag.train()
        swag(torch.randn(1, 1))
        swag.update_wrapper(0)
        assert swag.swag_stats["model.swag_stats.linear.weight_covariance_sqrt"].shape == (0, 10)
        swag.bn_update(dl, "cpu")
        swag(torch.randn(1, 1))

        swag.update_wrapper(1)
        assert swag.swag_stats["model.swag_stats.linear.weight_covariance_sqrt"].shape == (0, 10)
        assert swag.num_avgd_models == 0
        swag.bn_update(dl, "cpu")

        swag.update_wrapper(2)
        assert swag.swag_stats["model.swag_stats.linear.weight_covariance_sqrt"].shape == (1, 10)
        swag.bn_update(dl, "cpu")
        swag(torch.randn(1, 1))
        swag.update_wrapper(3)
        assert swag.swag_stats["model.swag_stats.linear.weight_covariance_sqrt"].shape == (2, 10)
        swag.update_wrapper(4)
        assert swag.num_avgd_models == 3
        assert swag.swag_stats["model.swag_stats.linear.weight_covariance_sqrt"].shape == (3, 10)
        swag.update_wrapper(5)
        assert swag.num_avgd_models == 4
        assert swag.swag_stats["model.swag_stats.linear.weight_covariance_sqrt"].shape == (3, 10)
        swag.eval()
        swag(torch.randn(1, 1))

        swag = SWAG(
            dummy_model(1, 10),
            cycle_start=1,
            cycle_length=1,
            diag_covariance=True,
        )
        swag.train()
        swag.update_wrapper(2)
        swag.sample(1, True, False, seed=1)

    def test_state_dict(self):
        mod = dummy_model(1, 10)
        swag = SWAG(mod, cycle_start=1, cycle_length=1, num_estimators=3)
        print(swag.state_dict())
        swag.load_state_dict(swag.state_dict())

    def test_failures(self):
        with pytest.raises(NotImplementedError, match="Raise an issue if you need this feature"):
            swag = SWAG(nn.Module(), scale=1, cycle_start=1, cycle_length=1)
            swag.sample(scale=1, block=True)
        with pytest.raises(ValueError, match="`scale` must be non-negative."):
            SWAG(nn.Module(), scale=-1, cycle_start=1, cycle_length=1)
        with pytest.raises(ValueError, match="`max_num_models` must be non-negative."):
            SWAG(nn.Module(), max_num_models=-1, cycle_start=1, cycle_length=1)
        with pytest.raises(ValueError, match="`var_clamp` must be non-negative. "):
            SWAG(nn.Module(), var_clamp=-1, cycle_start=1, cycle_length=1)
        swag = SWAG(nn.Module(), cycle_start=1, cycle_length=1, diag_covariance=True)
        with pytest.raises(
            ValueError,
            match="Cannot sample full rank from diagonal covariance matrix.",
        ):
            swag.sample(scale=1, diag_covariance=False)



# ruff: noqa: F401
from .adaptive_calibration_error import AdaptiveCalibrationError
from .brier_score import BrierScore
from .calibration_error import CalibrationError
from .categorical_nll import CategoricalNLL
from .disagreement import Disagreement
from .entropy import Entropy
from .fpr import FPR95, FPRx
from .grouping_loss import GroupingLoss
from .mean_iou import MeanIntersectionOverUnion
from .mutual_information import MutualInformation
from .risk_coverage import (
    AUGRC,
    AURC,
    CovAt5Risk,
    CovAtxRisk,
    RiskAt80Cov,
    RiskAtxCov,
)
from .variation_ratio import VariationRatio



from .base import ImageNetVariation


class ImageNetR(ImageNetVariation):
    url = "https://people.eecs.berkeley.edu/~hendrycks/imagenet-r.tar"
    filename = "imagenet-r.tar"
    tgz_md5 = "a61312130a589d0ca1a8fca1f2bd3337"
    dataset_name = "imagenet-r"



import logging
from collections.abc import Callable
from pathlib import Path

import numpy as np
import pandas as pd
import torch
from torchvision.datasets.utils import (
    download_and_extract_archive,
    extract_archive,
)

from .uci_classification import UCIClassificationDataset


class BankMarketing(UCIClassificationDataset):
    """The bank Marketing UCI classification dataset.

    Args:
        root (str): Root directory of the datasets.
        train (bool, optional): If True, creates dataset from training set,
            otherwise creates from test set.
        transform (callable, optional): A function/transform that takes in a
            numpy array and returns a transformed version.
        target_transform (callable, optional): A function/transform that takes
            in the target and transforms it.
        download (bool, optional): If true, downloads the dataset from the
            internet and puts it in root directory. If dataset is already
            downloaded, it is not downloaded again.
        binary (bool, optional): Whether to use binary classification. Defaults
            to ``True``.

    Note - License:
        The licenses of the datasets may differ from TorchUncertainty's
        license. Check before use.

    """

    md5_zip = "3a3c6c4189975ea1f3040dbd60ad106c"
    url = "https://archive.ics.uci.edu/static/public/222/bank+marketing.zip"
    dataset_name = "bank+marketing"
    filename = "bank-additional-full.csv"
    num_features = 62

    def __init__(
        self,
        root: Path | str,
        transform: Callable | None = None,
        target_transform: Callable | None = None,
        binary: bool = True,
        download: bool = False,
        train: bool = True,
        test_split: float = 0.2,
        split_seed: int = 21893027,
    ) -> None:
        super().__init__(
            root,
            transform,
            target_transform,
            binary,
            download,
            train,
            test_split,
            split_seed,
        )

    def download(self) -> None:
        """Download and extract dataset."""
        if self._check_integrity():
            logging.info("Files already downloaded and verified")
            return
        download_and_extract_archive(
            self.url,
            download_root=self.root,
            filename="bank+marketing.zip",
            md5=self.md5_zip,
        )
        extract_archive(self.root / "bank-additional.zip", self.root / "bank-marketing")

    def _make_dataset(self) -> None:
        """Create dataset from extracted files."""
        data = pd.read_csv(
            self.root / "bank-marketing" / "bank-additional" / self.filename,
            sep=";",
        )
        data["y"] = np.where(data["y"] == "yes", 1, 0)
        self.targets = torch.as_tensor(data["y"].values, dtype=torch.long)

        self.data = data.drop(columns=["y"])
        categorical_columns = self.data.select_dtypes(include="object").columns
        for col in categorical_columns:
            if self.data[col].nunique() == 2:
                self.data[col] = np.where(self.data[col] == "yes", 1, 0)
        self.data = torch.as_tensor(
            pd.get_dummies(self.data).astype(float).values, dtype=torch.float32
        )
        self.num_features = self.data.shape[1]



import torch
from torch import Tensor
from torchmetrics import MeanAbsoluteError


class Log10(MeanAbsoluteError):
    def __init__(self, **kwargs) -> None:
        r"""The Log10 metric.

        .. math:: \text{Log10} = \frac{1}{N} \sum_{i=1}^{N} |\log_{10}(y_i) - \log_{10}(\hat{y_i})|

        where :math:`N` is the batch size, :math:`y_i` is a tensor of target
            values and :math:`\hat{y_i}` is a tensor of prediction.

        Inputs:
            - :attr:`preds`: :math:`(N)`
            - :attr:`target`: :math:`(N)`

        Args:
            kwargs: Additional keyword arguments, see `Advanced metric settings
                <https://torchmetrics.readthedocs.io/en/stable/pages/overview.html#metric-kwargs>`_.
        """
        super().__init__(**kwargs)
        self.add_state("values", default=torch.tensor(0.0), dist_reduce_fx="sum")
        self.add_state("total", default=torch.tensor(0), dist_reduce_fx="sum")

    def update(self, pred: Tensor, target: Tensor) -> None:
        """Update state with predictions and targets."""
        return super().update(pred.log10(), target.log10())



from torch import Tensor, distributions
from torchmetrics.utilities.data import dim_zero_cat

from torch_uncertainty.metrics import CategoricalNLL


class DistributionNLL(CategoricalNLL):
    def update(
        self,
        dist: distributions.Distribution,
        target: Tensor,
        padding_mask: Tensor | None = None,
    ) -> None:
        """Update state with the predicted distributions and the targets.

        Args:
            dist (torch.distributions.Distribution): Predicted distributions.
            target (Tensor): Ground truth labels.
            padding_mask (Tensor, optional): The padding mask. Defaults to None.
                Sets the loss to 0 for padded values.
        """
        nlog_prob = -dist.log_prob(target)
        if padding_mask is not None:
            nlog_prob = nlog_prob.masked_fill(padding_mask, float("nan"))
        if self.reduction is None or self.reduction == "none":
            self.values.append(nlog_prob)
        else:
            self.values += nlog_prob.nansum()
            self.total += padding_mask.sum() if padding_mask is not None else target.numel()

    def compute(self) -> Tensor:
        """Computes NLL based on inputs passed in to ``update`` previously."""
        values = dim_zero_cat(self.values)

        if self.reduction == "sum":
            return values.nansum()
        if self.reduction == "mean":
            return values.nansum() / self.total
        # reduction is None or "none"
        return values



import copy

import torch
from torch import nn


class CheckpointEnsemble(nn.Module):
    def __init__(
        self,
        model: nn.Module,
        save_schedule: list[int] | None = None,
        use_final_checkpoint: bool = True,
    ) -> None:
        """Ensemble of models at different points in the training trajectory.

        Args:
            model (nn.Module): The model to train and ensemble.
            save_schedule (list[int]): The epochs at which to save the model.
                If save schedule is None, save the model at every epoch.
                Defaults to None.
            use_final_checkpoint (bool, optional): Whether to use the final
                model as a checkpoint. Defaults to True.

        Reference:
            Checkpoint Ensembles: Ensemble Methods from a Single Training Process.
            Hugh Chen, Scott Lundberg, Su-In Lee. In ArXiv 2018.
        """
        super().__init__()
        self.core_model = model
        self.save_schedule = save_schedule
        self.use_final_checkpoint = use_final_checkpoint
        self.num_estimators = int(use_final_checkpoint)
        self.saved_models = []
        self.num_estimators = 1

    @torch.no_grad()
    def update_wrapper(self, epoch: int) -> None:
        """Save the model at the given epoch if included in the schedule.

        Args:
            epoch (int): The current epoch.
        """
        if self.save_schedule is None or epoch in self.save_schedule:
            self.saved_models.append(copy.deepcopy(self.core_model))
            self.num_estimators += 1

    def eval_forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward pass for evaluation.

        If the model is in evaluation mode, this method will return the
        ensemble prediction. Otherwise, it will return the prediction of the
        current model.

        Args:
            x (torch.Tensor): The input tensor.

        Returns:
            torch.Tensor: The model or ensemble output.
        """
        if not len(self.saved_models):
            return self.core_model.forward(x)
        preds = torch.cat([model.forward(x) for model in self.saved_models], dim=0)
        if self.use_final_checkpoint:
            model_forward = self.core_model.forward(x)
            preds = torch.cat([model_forward, preds], dim=0)
        return preds

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        if self.training:
            return self.core_model.forward(x)
        return self.eval_forward(x)



import json
import logging
import shutil
from collections.abc import Callable
from pathlib import Path
from typing import Literal

from PIL import Image
from torchvision import tv_tensors
from torchvision.datasets import VisionDataset
from torchvision.datasets.utils import (
    download_and_extract_archive,
    download_url,
)
from torchvision.transforms import functional as F
from tqdm import tqdm


class KITTIDepth(VisionDataset):
    root: Path
    depth_url = "https://s3.eu-central-1.amazonaws.com/avg-kitti/data_depth_annotated.zip"
    depth_md5 = "7d1ce32633dc2f43d9d1656a1f875e47"
    raw_url = "https://s3.eu-central-1.amazonaws.com/avg-kitti/raw_data/"
    raw_filenames_url = "https://raw.githubusercontent.com/torch-uncertainty/dataset-metadata/main/download/kitti/raw_filenames.json"
    raw_filenames_md5 = "e5b7fad5ecd059488ef6c02dc9e444c1"
    _num_samples = {
        "train": 42949,
        "val": 3426,
        "test": ...,
    }

    def __init__(
        self,
        root: str | Path,
        split: Literal["train", "val"],
        min_depth: float = 0.0,
        max_depth: float = 80.0,
        transforms: Callable | None = None,
        download: bool = False,
        remove_unused: bool = False,
    ) -> None:
        logging.info(
            "KITTIDepth is copyrighted by the Karlsruhe Institute of Technology "
            "(KIT) and the Toyota Technological Institute at Chicago (TTIC). "
            "By using KITTIDepth, you agree to the terms and conditions of the "
            "Creative Commons Attribution-NonCommercial-ShareAlike 3.0 License. "
            "This means that you must attribute the work in the manner specified "
            "by the authors, you may not use this work for commercial purposes "
            "and if you alter, transform, or build upon this work, you may "
            "distribute the resulting work only under the same license."
        )

        super().__init__(
            root=Path(root) / "KITTIDepth",
            transforms=transforms,
        )
        self.min_depth = min_depth
        self.max_depth = max_depth

        if split not in ["train", "val"]:
            raise ValueError(f"split must be one of ['train', 'val']. Got {split}.")

        self.split = split

        if not self.check_split_integrity("leftDepth"):
            if download:
                self._download_depth()
            else:
                raise FileNotFoundError(
                    f"KITTI {split} split not found or incomplete. Set download=True to download it."
                )

        if not self.check_split_integrity("leftImg8bit"):
            if download:
                self._download_raw(remove_unused)
            else:
                raise FileNotFoundError(
                    f"KITTI {split} split not found or incomplete. Set download=True to download it."
                )

        self._make_dataset()

    def check_split_integrity(self, folder: str) -> bool:
        split_path = self.root / self.split
        return (
            split_path.is_dir()
            and len(list((split_path / folder).glob("*.png"))) == self._num_samples[self.split]
        )

    def __getitem__(self, index: int) -> tuple[tv_tensors.Image, tv_tensors.Mask]:
        """Get the sample at the given index.

        Args:
            index (int): Index

        Returns:
            tuple: (image, target) where target is a depth map.
        """
        image = tv_tensors.Image(Image.open(self.samples[index]).convert("RGB"))
        target = tv_tensors.Mask(
            F.pil_to_tensor(Image.open(self.targets[index])).squeeze(0) / 256.0
        )
        target[(target <= self.min_depth) | (target > self.max_depth)] = float("nan")

        if self.transforms is not None:
            image, target = self.transforms(image, target)

        return image, target

    def __len__(self) -> int:
        """The number of samples in the dataset."""
        return self._num_samples[self.split]

    def _make_dataset(self) -> None:
        self.samples = sorted((self.root / self.split / "leftImg8bit").glob("*.png"))
        self.targets = sorted((self.root / self.split / "leftDepth").glob("*.png"))

    def _download_depth(self) -> None:
        """Download and extract the depth annotation dataset."""
        if not (self.root / "tmp").exists():
            download_and_extract_archive(
                self.depth_url,
                download_root=self.root,
                extract_root=self.root / "tmp",
                md5=self.depth_md5,
            )

        logging.info("Re-structuring the depth annotations...")

        if (self.root / "train" / "leftDepth").exists():
            shutil.rmtree(self.root / "train" / "leftDepth")

        (self.root / "train" / "leftDepth").mkdir(parents=True, exist_ok=False)

        depth_files = list((self.root).glob("**/tmp/train/**/image_02/*.png"))
        logging.info("Train files...")
        for file in tqdm(depth_files):
            exp_code = file.parents[3].name.split("_")
            filecode = "_".join([exp_code[0], exp_code[1], exp_code[2], exp_code[4], file.name])
            shutil.copy(file, self.root / "train" / "leftDepth" / filecode)

        if (self.root / "val" / "leftDepth").exists():
            shutil.rmtree(self.root / "val" / "leftDepth")

        (self.root / "val" / "leftDepth").mkdir(parents=True, exist_ok=False)

        depth_files = list((self.root).glob("**/tmp/val/**/image_02/*.png"))
        logging.info("Validation files...")
        for file in tqdm(depth_files):
            exp_code = file.parents[3].name.split("_")
            filecode = "_".join([exp_code[0], exp_code[1], exp_code[2], exp_code[4], file.name])
            shutil.copy(file, self.root / "val" / "leftDepth" / filecode)

        shutil.rmtree(self.root / "tmp")

    def _download_raw(self, remove_unused: bool) -> None:
        """Download and extract the raw dataset."""
        download_url(
            self.raw_filenames_url,
            self.root,
            "raw_filenames.json",
            self.raw_filenames_md5,
        )
        with (self.root / "raw_filenames.json").open() as file:
            raw_filenames = json.load(file)

        for filename in tqdm(raw_filenames):
            logging.info("%s", self.raw_url + filename)
            download_and_extract_archive(
                self.raw_url + filename,
                download_root=self.root,
                extract_root=self.root / "raw",
                md5=None,
            )

        logging.info("Re-structuring the raw data...")

        samples_to_keep = list((self.root / "train" / "leftDepth").glob("*.png"))

        if (self.root / "train" / "leftImg8bit").exists():
            shutil.rmtree(self.root / "train" / "leftImg8bit")

        (self.root / "train" / "leftImg8bit").mkdir(parents=True, exist_ok=False)

        logging.info("Train files...")
        for sample in tqdm(samples_to_keep):
            filecode = sample.name.split("_")
            first_level = "_".join([filecode[0], filecode[1], filecode[2]])
            second_level = "_".join(
                [
                    filecode[0],
                    filecode[1],
                    filecode[2],
                    "drive",
                    filecode[3],
                    "sync",
                ]
            )
            raw_path = (
                self.root / "raw" / first_level / second_level / "image_02" / "data" / filecode[4]
            )
            shutil.copy(raw_path, self.root / "train" / "leftImg8bit" / sample.name)

        samples_to_keep = list((self.root / "val" / "leftDepth").glob("*.png"))

        if (self.root / "val" / "leftImg8bit").exists():
            shutil.rmtree(self.root / "val" / "leftImg8bit")

        (self.root / "val" / "leftImg8bit").mkdir(parents=True, exist_ok=False)

        logging.info("Validation files...")
        for sample in tqdm(samples_to_keep):
            filecode = sample.name.split("_")
            first_level = "_".join([filecode[0], filecode[1], filecode[2]])
            second_level = "_".join(
                [
                    filecode[0],
                    filecode[1],
                    filecode[2],
                    "drive",
                    filecode[3],
                    "sync",
                ]
            )
            raw_path = (
                self.root / "raw" / first_level / second_level / "image_02" / "data" / filecode[4]
            )
            shutil.copy(raw_path, self.root / "val" / "leftImg8bit" / sample.name)

        if remove_unused:
            shutil.rmtree(self.root / "raw")



import pytest
import torch

from torch_uncertainty.models.segmentation.deeplab import (
    _DeepLabV3,
    deep_lab_v3_resnet50,
    deep_lab_v3_resnet101,
)


class TestDeeplab:
    """Testing the Deeplab class."""

    @torch.no_grad()
    def test_main(self):
        model = deep_lab_v3_resnet50(10, "v3", 16, True, False).eval()
        model(torch.randn(1, 3, 32, 32))
        model = deep_lab_v3_resnet50(10, "v3", 16, False, False).eval()
        model = deep_lab_v3_resnet101(10, "v3+", 8, True, False).eval()
        model(torch.randn(1, 3, 32, 32))
        model = deep_lab_v3_resnet101(10, "v3+", 8, False, False).eval()

    def test_errors(self):
        with pytest.raises(ValueError, match="Unknown backbone:"):
            _DeepLabV3(10, "other", "v3", 16, True, False)
        with pytest.raises(ValueError, match="output_stride: "):
            deep_lab_v3_resnet50(10, "v3", 15, True, False)
        with pytest.raises(ValueError, match="Unknown style: "):
            deep_lab_v3_resnet50(10, "v2", 16, True, False)



import pytest
import torch
from torch import nn

from tests._dummies.model import dummy_model
from torch_uncertainty.models import EMA


class TestEMA:
    """Testing the EMA class."""

    def test_training(self):
        ema = EMA(dummy_model(1, 10), momentum=0.99)
        ema.eval()
        ema(torch.randn(1, 1))
        ema.train()
        ema.update_wrapper(0)

    def test_failures(self):
        with pytest.raises(ValueError, match="must be in the range"):
            EMA(nn.Module(), momentum=-1)



import pytest
import torch

from torch_uncertainty.losses import (
    BCEWithLogitsLSLoss,
    ConfidencePenaltyLoss,
    ConflictualLoss,
    DECLoss,
    FocalLoss,
)


class TestDECLoss:
    """Testing the DECLoss class."""

    def test_main(self):
        loss = DECLoss(loss_type="mse", reg_weight=1e-2, annealing_step=1, reduction="sum")
        loss(torch.tensor([[0.0, 0.0]]), torch.tensor([0]), current_epoch=1)
        loss = DECLoss(loss_type="mse", reg_weight=1e-2, annealing_step=1)
        loss(torch.tensor([[0.0, 0.0]]), torch.tensor([0]), current_epoch=0)
        loss = DECLoss(loss_type="log", reg_weight=1e-2, reduction="none")
        loss(torch.tensor([[0.0, 0.0]]), torch.tensor([0]))
        loss = DECLoss(loss_type="digamma")
        loss(torch.tensor([[0.0, 0.0]]), torch.tensor([0]))

    def test_failures(self):
        with pytest.raises(
            ValueError,
            match="The regularization weight should be non-negative, but got",
        ):
            DECLoss(reg_weight=-1)

        with pytest.raises(ValueError, match="The annealing step should be positive, but got "):
            DECLoss(annealing_step=0)

        loss = DECLoss(annealing_step=10)
        with pytest.raises(ValueError):
            loss(
                torch.tensor([[0.0, 0.0]]),
                torch.tensor([0]),
                current_epoch=None,
            )

        with pytest.raises(ValueError, match=" is not a valid value for reduction."):
            DECLoss(reduction="median")

        with pytest.raises(ValueError, match="is not a valid value for mse/log/digamma loss."):
            DECLoss(loss_type="regression")


class TestConfidencePenaltyLoss:
    """Testing the ConfidencePenaltyLoss class."""

    def test_main(self):
        loss = ConfidencePenaltyLoss(reg_weight=1e-2, reduction="sum")
        loss(torch.tensor([[0.0, 0.0]]), torch.tensor([0]))
        loss = ConfidencePenaltyLoss(reg_weight=1e-2)
        loss(torch.tensor([[0.0, 0.0]]), torch.tensor([0]))
        loss = ConfidencePenaltyLoss(reg_weight=1e-2, reduction=None)
        loss(torch.tensor([[0.0, 0.0]]), torch.tensor([0]))

    def test_failures(self):
        with pytest.raises(
            ValueError,
            match="The regularization weight should be non-negative, but got",
        ):
            ConfidencePenaltyLoss(reg_weight=-1)

        with pytest.raises(ValueError, match="is not a valid value for reduction."):
            ConfidencePenaltyLoss(reduction="median")

        with pytest.raises(
            ValueError,
            match="The epsilon value should be non-negative, but got",
        ):
            ConfidencePenaltyLoss(eps=-1)


class TestConflictualLoss:
    """Testing the ConflictualLoss class."""

    def test_main(self):
        loss = ConflictualLoss(reg_weight=1e-2, reduction="sum")
        loss(torch.tensor([[0.0, 0.0]]), torch.tensor([0]))
        loss = ConflictualLoss(reg_weight=1e-2)
        loss(torch.tensor([[0.0, 0.0]]), torch.tensor([0]))
        loss = ConflictualLoss(reg_weight=1e-2, reduction=None)
        loss(torch.tensor([[0.0, 0.0]]), torch.tensor([0]))

    def test_failures(self):
        with pytest.raises(
            ValueError,
            match="The regularization weight should be non-negative, but got",
        ):
            ConflictualLoss(reg_weight=-1)

        with pytest.raises(ValueError, match="is not a valid value for reduction."):
            ConflictualLoss(reduction="median")


class TestFocalLoss:
    """Testing the FocalLoss class."""

    def test_main(self):
        loss = FocalLoss(gamma=1, reduction="sum")
        loss(torch.tensor([[0.0, 0.0]]), torch.tensor([0]))
        loss = FocalLoss(gamma=0.5)
        loss(torch.tensor([[0.0, 0.0]]), torch.tensor([0]))
        loss = FocalLoss(gamma=0.5, reduction=None)
        loss(torch.tensor([[0.0, 0.0]]), torch.tensor([0]))

    def test_failures(self):
        with pytest.raises(
            ValueError,
            match="The gamma term of the focal loss should be non-negative, but got",
        ):
            FocalLoss(gamma=-1)

        with pytest.raises(ValueError, match="is not a valid value for reduction."):
            FocalLoss(gamma=1, reduction="median")


class TestBCEWithLogitsLSLoss:
    """Testing the BCEWithLogitsLSLoss class."""

    def test_main(self):
        loss = BCEWithLogitsLSLoss(reduction="sum", label_smoothing=0.1, weight=torch.Tensor([1]))
        loss(torch.tensor([0.0]), torch.tensor([0]))
        loss = BCEWithLogitsLSLoss(reduction="mean", label_smoothing=0.6)
        loss(torch.tensor([0.0]), torch.tensor([0]))
        loss = BCEWithLogitsLSLoss(reduction="none", label_smoothing=0.1)
        loss(torch.tensor([0.0]), torch.tensor([0]))
        loss = BCEWithLogitsLSLoss(reduction="none")
        loss(torch.tensor([0.0]), torch.tensor([0]))



"""Adapted from https://github.com/hendrycks/robustness."""

from importlib import util
from io import BytesIO

if util.find_spec("cv2"):
    import cv2

    cv2_installed = True
else:  # coverage: ignore
    cv2_installed = False

import numpy as np
import torch
from PIL import Image

if util.find_spec("skimage"):
    from skimage.filters import gaussian
    from skimage.util import random_noise

    skimage_installed = True
else:  # coverage: ignore
    skimage_installed = False

if util.find_spec("scipy"):
    from scipy.ndimage import map_coordinates
    from scipy.ndimage import zoom as scizoom

    scipy_installed = True
else:  # coverage: ignore
    scipy_installed = False

from torch import Tensor, nn
from torchvision.transforms import (
    InterpolationMode,
    RandomResizedCrop,
    Resize,
    ToPILImage,
    ToTensor,
)

if util.find_spec("kornia"):
    from kornia.filters import motion_blur

    kornia_installed = True
else:  # coverage: ignore
    kornia_installed = False

from torch_uncertainty.datasets import FrostImages

from .image import Brightness as IBrightness
from .image import Contrast as IContrast
from .image import Saturation as ISaturation

__all__ = [
    "Brightness",
    "Contrast",
    "DefocusBlur",
    "Elastic",
    "Fog",
    "Frost",
    "GaussianBlur",
    "GaussianNoise",
    "GlassBlur",
    "ImpulseNoise",
    "JPEGCompression",
    "MotionBlur",
    "Pixelate",
    "Saturation",
    "ShotNoise",
    "Snow",
    "SpeckleNoise",
    "ZoomBlur",
    "corruption_transforms",
]


class TUCorruption(nn.Module):
    def __init__(self, severity: int) -> None:
        """Base class for corruptions."""
        super().__init__()
        if not (0 <= severity <= 5):
            raise ValueError("Severity must be between 0 and 5.")
        if not isinstance(severity, int):
            raise TypeError("Severity must be an integer.")
        self.severity = severity

    def __repr__(self) -> str:
        """Printable representation."""
        return self.__class__.__name__ + f"(severity={self.severity})"


class GaussianNoise(TUCorruption):
    def __init__(self, severity: int) -> None:
        """Add Gaussian noise to an image.

        Args:
            severity (int): Severity level of the corruption.
        """
        super().__init__(severity)
        self.scale = [0, 0.04, 0.06, 0.08, 0.09, 0.10][severity]

    def forward(self, img: Tensor) -> Tensor:
        if self.severity == 0:
            return img
        return torch.clamp(torch.normal(img, self.scale), 0, 1)


class ShotNoise(TUCorruption):
    def __init__(self, severity: int) -> None:
        """Add shot noise to an image.

        Args:
            severity (int): Severity level of the corruption.
        """
        super().__init__(severity)
        self.scale = [500, 250, 100, 75, 50][severity - 1]

    def forward(self, img: Tensor):
        if self.severity == 0:
            return img
        return torch.clamp(torch.poisson(img * self.scale) / self.scale, 0, 1)


class ImpulseNoise(TUCorruption):
    def __init__(self, severity: int) -> None:
        """Add impulse noise to an image.

        Args:
            severity (int): Severity level of the corruption.
        """
        super().__init__(severity)
        if not skimage_installed:
            raise ImportError(
                "Please install torch_uncertainty with the image option:"
                """pip install -U "torch_uncertainty[image]"."""
            )
        self.scale = [0, 0.01, 0.02, 0.03, 0.05, 0.07][severity]

    def forward(self, img: Tensor) -> Tensor:
        if self.severity == 0:
            return img
        return torch.clamp(
            torch.as_tensor(random_noise(img, mode="s&p", amount=self.scale)),
            torch.zeros(1),
            torch.ones(1),
        )


class DefocusBlur(TUCorruption):
    def __init__(self, severity: int) -> None:
        """Add defocus blur to an image.

        Args:
            severity (int): Severity level of the corruption.
        """
        super().__init__(severity)
        if not cv2_installed:
            raise ImportError(
                "Please install torch_uncertainty with the image option:"
                """pip install -U "torch_uncertainty[image]"."""
            )
        self.radius = [3, 4, 6, 8, 10][severity - 1]
        self.alias_blur = [0.1, 0.5, 0.5, 0.5, 0.5][severity - 1]

    def forward(self, img: Tensor) -> Tensor:
        if self.severity == 0:
            return img
        img = img.numpy()
        channels = [
            torch.as_tensor(
                cv2.filter2D(
                    img[ch, :, :],
                    -1,
                    disk(self.radius, alias_blur=self.alias_blur),
                )
            )
            for ch in range(3)
        ]
        return torch.clamp(torch.stack(channels), 0, 1)


class GlassBlur(TUCorruption):  # TODO: batch
    def __init__(self, severity: int) -> None:
        super().__init__(severity)
        if not skimage_installed or not cv2_installed:
            raise ImportError(
                "Please install torch_uncertainty with the image option:"
                """pip install -U "torch_uncertainty[image]"."""
            )
        self.sigma = [0.05, 0.25, 0.4, 0.25, 0.4][severity - 1]
        self.max_delta = 1
        self.iterations = [1, 1, 1, 2, 2][severity - 1]

    def forward(self, img: Tensor) -> Tensor:
        img_size = img.shape
        img = torch.as_tensor(gaussian(img, sigma=self.sigma))
        for _ in range(self.iterations):
            for h in range(img_size[0] - self.max_delta, self.max_delta, -1):
                for w in range(img_size[1] - self.max_delta, self.max_delta, -1):
                    dx, dy = torch.randint(-self.max_delta, self.max_delta, size=(2,))
                    h_prime, w_prime = h + dy, w + dx
                    img[h, w], img[h_prime, w_prime] = (
                        img[h_prime, w_prime],
                        img[h, w],
                    )
        return torch.clamp(torch.as_tensor(gaussian(img, sigma=self.sigma)), 0, 1)


def disk(radius: int, alias_blur: float = 0.1, dtype=np.float32):
    if radius <= 8:
        size = np.arange(-8, 8 + 1)
        ksize = (3, 3)
    else:  # coverage: ignore
        size = np.arange(-radius, radius + 1)
        ksize = (5, 5)
    xs, ys = np.meshgrid(size, size)
    aliased_disk = np.array((xs**2 + ys**2) <= radius**2, dtype=dtype)
    aliased_disk /= np.sum(aliased_disk)
    return cv2.GaussianBlur(aliased_disk, ksize=ksize, sigmaX=alias_blur)


class MotionBlur(TUCorruption):
    def __init__(self, severity: int) -> None:
        """Apply a motion blur corruption on the image.

        Note:
            Originally, Hendrycks et al. used Gaussian motion blur. To remove the dependency with
            with `Wand` we changed the transform to a simpler motion blur and kept the values of
            sigma as the new half kernel sizes.
        """
        super().__init__(severity)
        self.rng = np.random.default_rng()
        self.radius = [3, 5, 8, 12, 15][severity - 1]

        if not kornia_installed:
            raise ImportError(
                "Please install torch_uncertainty with the image option:"
                """pip install -U "torch_uncertainty[image]"."""
            )

    def forward(self, img: Tensor) -> Tensor:
        if self.severity == 0:
            return img
        no_batch = False
        if img.ndim == 3:
            no_batch = True
            img = img.unsqueeze(0)
        out = motion_blur(
            img, kernel_size=self.radius * 2 + 1, angle=self.rng.uniform(-45, 45), direction=0
        )
        if no_batch:
            out = out.squeeze(0)
        return out


def clipped_zoom(img, zoom_factor):
    h = img.shape[0]
    # ceil crop height(= crop width)
    ch = int(np.ceil(h / zoom_factor))

    top = (h - ch) // 2
    img = scizoom(
        img[top : top + ch, top : top + ch],
        (zoom_factor, zoom_factor, 1),
        order=1,
    )
    # trim off any extra pixels
    trim_top = (img.shape[0] - h) // 2

    return img[trim_top : trim_top + h, trim_top : trim_top + h]


class ZoomBlur(TUCorruption):
    def __init__(self, severity: int) -> None:
        super().__init__(severity)
        self.zooms = [
            np.arange(1, 1.11, 0.01),
            np.arange(1, 1.16, 0.01),
            np.arange(1, 1.21, 0.02),
            np.arange(1, 1.26, 0.02),
            np.arange(1, 1.31, 0.03),
        ][severity - 1]

        if not scipy_installed:
            raise ImportError(
                "Please install torch_uncertainty with the all option:"
                """pip install -U "torch_uncertainty[all]"."""
            )

    def forward(self, img: Tensor) -> Tensor:
        if self.severity == 0:
            return img
        img = img.permute(1, 2, 0).numpy()
        out = np.zeros_like(img)
        for zoom_factor in self.zooms:
            out += clipped_zoom(img, zoom_factor)
        img = (img + out) / (len(self.zooms) + 1)
        return torch.clamp(torch.as_tensor(img).permute(2, 0, 1), 0, 1)


class Snow(TUCorruption):
    def __init__(self, severity: int) -> None:
        """Apply a snow effect on the image.

        Note:
            The transformation has been slightly modified, see MotionBlur for details.
        """
        super().__init__(severity)
        self.mix = [
            (0.1, 0.3, 3, 0.5, 4, 0.8),
            (0.2, 0.3, 2, 0.5, 4, 0.7),
            (0.55, 0.3, 4, 0.9, 8, 0.7),
            (0.55, 0.3, 4.5, 0.85, 8, 0.65),
            (0.55, 0.3, 2.5, 0.85, 12, 0.55),
        ][severity - 1]
        self.rng = np.random.default_rng()

        if not kornia_installed:
            raise ImportError(
                "Please install torch_uncertainty with the image option:"
                """pip install -U "torch_uncertainty[image]"."""
            )

    def forward(self, img: Tensor) -> Tensor:
        if self.severity == 0:
            return img
        _, height, width = img.shape
        x = img.numpy()
        snow_layer = self.rng.normal(size=x.shape[1:], loc=self.mix[0], scale=self.mix[1])[
            ..., np.newaxis
        ]
        snow_layer = clipped_zoom(snow_layer, self.mix[2])
        snow_layer[snow_layer < self.mix[3]] = 0
        snow_layer = np.clip(snow_layer.squeeze(), 0, 1)

        snow_layer = (
            motion_blur(
                torch.as_tensor(snow_layer).unsqueeze(0).unsqueeze(0),
                kernel_size=self.mix[4] * 2 + 1,
                angle=self.rng.uniform(-135, -45),
                direction=0,
            )
            .squeeze(0)
            .numpy()
        )

        x = self.mix[5] * x + (1 - self.mix[5]) * np.maximum(
            x,
            cv2.cvtColor(x.transpose([1, 2, 0]), cv2.COLOR_RGB2GRAY).reshape(1, height, width) * 1.5
            + 0.5,
        )
        return torch.clamp(torch.as_tensor(x + snow_layer + np.rot90(snow_layer, k=2)), 0, 1)


class Frost(TUCorruption):
    def __init__(self, severity: int) -> None:
        super().__init__(severity)
        self.rng = np.random.default_rng()
        self.mix = [(1, 0.2), (1, 0.3), (0.9, 0.4), (0.85, 0.4), (0.75, 0.45)][severity - 1]
        self.frost_ds = FrostImages("./data", download=True, transform=ToTensor())

    def forward(self, img: Tensor) -> Tensor:
        if self.severity == 0:
            return img
        _, height, width = img.shape
        frost_img = RandomResizedCrop((height, width))(
            self.frost_ds[self.rng.integers(low=0, high=4)]
        )
        return torch.clamp(self.mix[0] * img + self.mix[1] * frost_img, 0, 1)


def plasma_fractal(height, width, wibbledecay=3):
    """Generate a heightmap using diamond-square algorithm.
    Return square 2d array, side length 'mapsize', of floats in range 0-1.
    'mapsize' must be a power of two.
    """
    maparray = np.empty((height, width), dtype=np.float64)
    maparray[0, 0] = 0
    stepsize = height
    wibble = 100
    rng = np.random.default_rng()

    def wibbledmean(array):
        return array / 4 + wibble * rng.uniform(-wibble, wibble, array.shape)

    def fillsquares():
        """For each square of points stepsize apart, calculate middle value as mean of points + wibble."""
        cornerref = maparray[0:height:stepsize, 0:height:stepsize]
        squareaccum = cornerref + np.roll(cornerref, shift=-1, axis=0)
        squareaccum += np.roll(squareaccum, shift=-1, axis=1)
        maparray[
            stepsize // 2 : height : stepsize,
            stepsize // 2 : height : stepsize,
        ] = wibbledmean(squareaccum)

    def filldiamonds():
        """For each diamond of points stepsize apart, calculate middle value as mean of points + wibble."""
        mapsize = maparray.shape[0]
        drgrid = maparray[
            stepsize // 2 : mapsize : stepsize,
            stepsize // 2 : mapsize : stepsize,
        ]
        ulgrid = maparray[0:mapsize:stepsize, 0:mapsize:stepsize]
        ldrsum = drgrid + np.roll(drgrid, 1, axis=0)
        lulsum = ulgrid + np.roll(ulgrid, -1, axis=1)
        ltsum = ldrsum + lulsum
        maparray[0:mapsize:stepsize, stepsize // 2 : mapsize : stepsize] = wibbledmean(ltsum)
        tdrsum = drgrid + np.roll(drgrid, 1, axis=1)
        tulsum = ulgrid + np.roll(ulgrid, -1, axis=0)
        ttsum = tdrsum + tulsum
        maparray[stepsize // 2 : mapsize : stepsize, 0:mapsize:stepsize] = wibbledmean(ttsum)

    while stepsize >= 2:
        fillsquares()
        filldiamonds()
        stepsize //= 2
        wibble /= wibbledecay

    maparray -= maparray.min()
    return maparray / maparray.max()


class Fog(TUCorruption):
    def __init__(self, severity: int, size: int = 256) -> None:
        super().__init__(severity)
        if (size & (size - 1) == 0) and size != 0:
            self.size = size
            self.resize = Resize((size, size), InterpolationMode.BICUBIC)
        else:
            raise ValueError(f"Size must be a power of 2. Got {size}.")
        self.mix = [(1.5, 2), (2, 2), (2.5, 1.7), (2.5, 1.5), (3, 1.4)][severity - 1]

    def forward(self, img: Tensor) -> Tensor:
        if self.severity == 0:
            return img
        _, height, width = img.shape
        if height != width:
            raise ValueError(f"Image must be square. Got {height}x{width}.")
        img = self.resize(img)
        max_val = img.max()
        fog = (
            self.mix[0]
            * plasma_fractal(height=height, width=width, wibbledecay=self.mix[1])[:height, :width]
        )
        final = torch.clamp((img + fog) * max_val / (max_val + self.mix[0]), 0, 1)
        return Resize((height, width), InterpolationMode.BICUBIC)(final)


class Brightness(IBrightness, TUCorruption):
    def __init__(self, severity: int) -> None:
        TUCorruption.__init__(self, severity)
        self.level = [1.1, 1.2, 1.3, 1.4, 1.5][severity - 1]

    def forward(self, img: Tensor) -> Tensor:
        if self.severity == 0:
            return img
        return IBrightness.forward(self, img, self.level)


class Contrast(IContrast, TUCorruption):
    def __init__(self, severity: int) -> None:
        TUCorruption.__init__(self, severity)
        self.level = [0.4, 0.3, 0.2, 0.1, 0.05][severity - 1]

    def forward(self, img: Tensor) -> Tensor | Image.Image:
        if self.severity == 0:
            return img
        return IContrast.forward(self, img, self.level)


class Pixelate(TUCorruption):
    def __init__(self, severity: int) -> None:
        super().__init__(severity)
        self.quality = [0.95, 0.9, 0.85, 0.75, 0.65][severity - 1]

    def forward(self, img: Tensor) -> Tensor:
        if self.severity == 0:
            return img
        _, height, width = img.shape
        img = ToPILImage()(img)
        img = Resize(
            (int(height * self.quality), int(width * self.quality)),
            InterpolationMode.BOX,
        )(img)
        return ToTensor()(Resize((height, width), InterpolationMode.BOX)(img))


class JPEGCompression(TUCorruption):
    def __init__(self, severity: int) -> None:
        super().__init__(severity)
        self.quality = [80, 65, 58, 50, 40][severity - 1]

    def forward(self, img: Tensor) -> Tensor:
        if self.severity == 0:
            return img
        output = BytesIO()
        ToPILImage()(img).save(output, "JPEG", quality=self.quality)
        return ToTensor()(Image.open(output))


class Elastic(TUCorruption):
    def __init__(self, severity: int) -> None:
        super().__init__(severity)
        if not cv2_installed or not scipy_installed:
            raise ImportError(
                "Please install torch_uncertainty with the all option:"
                """pip install -U "torch_uncertainty[all]"."""
            )
        # The following pertubation values are based on the original repo but
        # are quite strange, notably for the severities 3 and 4
        self.mix = [
            (2, 0.7, 0.1),
            (2, 0.08, 0.2),
            (0.05, 0.01, 0.02),
            (0.07, 0.01, 0.02),
            (0.12, 0.01, 0.02),
        ][severity - 1]
        self.rng = np.random.default_rng()

    def forward(self, img: Tensor) -> Tensor:
        if self.severity == 0:
            return img
        image = np.array(img.permute(1, 2, 0), dtype=np.float32)
        shape = image.shape
        shape_size = shape[:2]

        # random affine
        center_square = np.float32(shape_size) // 2
        square_size = min(shape_size) // 3
        pts1 = np.float32(
            [
                center_square + square_size,
                [
                    center_square[0] + square_size,
                    center_square[1] - square_size,
                ],
                center_square - square_size,
            ]
        )
        pts2 = pts1 + self.rng.uniform(
            -self.mix[2] * shape_size[0],
            self.mix[2] * shape_size[0],
            size=pts1.shape,
        ).astype(np.float32)
        affine_transform = cv2.getAffineTransform(pts1, pts2)
        image = cv2.warpAffine(
            image,
            affine_transform,
            shape_size[::-1],
            borderMode=cv2.BORDER_REFLECT_101,
        )

        dx = (
            gaussian(
                self.rng.uniform(-1, 1, size=shape[:2]),
                self.mix[1] * shape_size[0],
                mode="reflect",
                truncate=3,
            )
            * self.mix[0]
            * shape_size[0]
        ).astype(np.float32)
        dy = (
            gaussian(
                self.rng.uniform(-1, 1, size=shape[:2]),
                self.mix[1] * shape_size[0],
                mode="reflect",
                truncate=3,
            )
            * self.mix[0]
            * shape_size[0]
        ).astype(np.float32)
        dx, dy = dx[..., np.newaxis], dy[..., np.newaxis]

        x, y, z = np.meshgrid(np.arange(shape[1]), np.arange(shape[0]), np.arange(shape[2]))
        indices = (
            np.reshape(y + dy, (-1, 1)),
            np.reshape(x + dx, (-1, 1)),
            np.reshape(z, (-1, 1)),
        )
        img = np.clip(
            map_coordinates(image, indices, order=1, mode="reflect").reshape(shape),
            0,
            1,
        )
        return torch.as_tensor(img).permute(2, 0, 1)


class SpeckleNoise(TUCorruption):
    def __init__(self, severity: int) -> None:
        super().__init__(severity)
        self.scale = [0.06, 0.1, 0.12, 0.16, 0.2][severity - 1]
        self.rng = np.random.default_rng()

    def forward(self, img: Tensor) -> Tensor:
        if self.severity == 0:
            return img
        return torch.clamp(
            img + img * self.rng.normal(img, self.scale),
            0,
            1,
        )


class GaussianBlur(TUCorruption):
    def __init__(self, severity: int) -> None:
        super().__init__(severity)
        if not skimage_installed:
            raise ImportError(
                "Please install torch_uncertainty with the image option:"
                """pip install -U "torch_uncertainty[image]"."""
            )
        self.sigma = [0.4, 0.6, 0.7, 0.8, 1.0][severity - 1]

    def forward(self, img: Tensor) -> Tensor:
        if self.severity == 0:
            return img
        return torch.clamp(
            torch.as_tensor(gaussian(img, sigma=self.sigma)),
            min=0,
            max=1,
        )


class Saturation(ISaturation, TUCorruption):
    def __init__(self, severity: int) -> None:
        TUCorruption.__init__(self, severity)
        self.severity = severity
        self.level = [0.1, 0.2, 0.3, 0.4, 0.5][severity - 1]

    def forward(self, img: Tensor) -> Tensor:
        if self.severity == 0:
            return img
        return ISaturation.forward(self, img, self.level)


corruption_transforms = (
    GaussianNoise,
    ShotNoise,
    ImpulseNoise,
    DefocusBlur,
    GlassBlur,
    MotionBlur,
    ZoomBlur,
    Snow,
    Frost,
    Fog,
    Brightness,
    Contrast,
    Elastic,
    Pixelate,
    JPEGCompression,
)



import matplotlib.pyplot as plt
import numpy as np
import torchvision.transforms.functional as F
from torch import Tensor


def show(prediction: Tensor, target: Tensor):
    imgs = [prediction, target]
    fig, axs = plt.subplots(ncols=len(imgs), figsize=(12, 6))
    for i, img in enumerate(imgs):
        img = img.detach()
        img = F.to_pil_image(img)
        axs[i].imshow(np.asarray(img))
        axs[i].set(xticklabels=[], yticklabels=[], xticks=[], yticks=[])

    axs[0].set(title="Prediction")
    axs[1].set(title="Ground Truth")

    return fig



import torch
from torch import Tensor
from torchmetrics import MeanAbsoluteError, MeanSquaredError


class MeanGTRelativeAbsoluteError(MeanAbsoluteError):
    def __init__(self, **kwargs) -> None:
        r"""Compute Mean Absolute Error relative to the Ground Truth (MAErel
            or ARErel).

        .. math:: \text{MAErel} = \frac{1}{N}\sum_i^N \frac{| y_i - \hat{y_i} |}{y_i}

        where :math:`y` is a tensor of target values, and :math:`\hat{y}` is a
            tensor of predictions.

        As input to ``forward`` and ``update`` the metric accepts the following
            input:

        - ``preds`` (:class:`~torch.Tensor`): Predictions from model
        - ``target`` (:class:`~torch.Tensor`): Ground truth values

        As output of ``forward`` and ``compute`` the metric returns the
            following output:

        - ``rel_mean_absolute_error`` (:class:`~torch.Tensor`): A tensor with
            the relative mean absolute error over the state

        Args:
            kwargs: Additional keyword arguments, see `Advanced metric settings
                <https://torchmetrics.readthedocs.io/en/stable/pages/overview.html#metric-kwargs>`_.

        Reference:
            As in e.g. From big to small: Multi-scale local planar guidance for
            monocular depth estimation
        """
        super().__init__(**kwargs)

    def update(self, pred: Tensor, target: Tensor) -> None:
        """Update state with predictions and targets."""
        return super().update(pred / target, torch.ones_like(target))


class MeanGTRelativeSquaredError(MeanSquaredError):
    def __init__(self, squared: bool = True, num_outputs: int = 1, **kwargs) -> None:
        r"""Compute mean squared error relative to the Ground Truth (MSErel or SRE).

        .. math:: \text{MSErel} = \frac{1}{N}\sum_i^N \frac{(y_i - \hat{y_i})^2}{y_i}

        Where :math:`y` is a tensor of target values, and :math:`\hat{y}` is a
            tensor of predictions.

        As input to ``forward`` and ``update`` the metric accepts the following
            input:

        - ``preds`` (:class:`~torch.Tensor`): Predictions from model
        - ``target`` (:class:`~torch.Tensor`): Ground truth values

        As output of ``forward`` and ``compute`` the metric returns the
            following output:

        - ``rel_mean_squared_error`` (:class:`~torch.Tensor`): A tensor with
            the relative mean squared error

        Args:
            squared: If True returns MSErel value, if False returns RMSErel
                value.
            num_outputs: Number of outputs in multioutput setting
            kwargs: Additional keyword arguments, see `Advanced metric settings
                <https://torchmetrics.readthedocs.io/en/stable/pages/overview.html#metric-kwargs>`_.

        Reference:
            As in e.g. From big to small: Multi-scale local planar guidance for
                monocular depth estimation
        """
        super().__init__(squared, num_outputs, **kwargs)

    def update(self, pred: Tensor, target: Tensor) -> None:
        """Update state with predictions and targets."""
        return super().update(pred / torch.sqrt(target), torch.sqrt(target))



def get_resnet_num_blocks(arch: int) -> list[int]:
    if arch == 18:
        num_blocks = [2, 2, 2, 2]
    elif arch == 20:
        num_blocks = [3, 3, 3]
    elif arch == 34 or arch == 50:
        num_blocks = [3, 4, 6, 3]
    elif arch == 44:
        num_blocks = [7, 7, 7]
    elif arch == 56:
        num_blocks = [9, 9, 9]
    elif arch == 101:
        num_blocks = [3, 4, 23, 3]
    elif arch == 110:
        num_blocks = [18, 18, 18]
    elif arch == 152:
        num_blocks = [3, 8, 36, 3]
    elif arch == 1202:
        num_blocks = [200, 200, 200]
    else:
        raise ValueError(f"Unknown ResNet architecture. Got {arch}.")
    return num_blocks



from torch import Tensor
from torchmetrics import MeanSquaredError


class MeanSquaredLogError(MeanSquaredError):
    def __init__(self, squared: bool = True, **kwargs) -> None:
        r"""MeanSquaredLogError (MSELog) regression metric.

        .. math:: \text{MSELog} = \frac{1}{N}\sum_i^N  (\log \hat{y_i} - \log y_i)^2

        where :math:`y` is a tensor of target values, and :math:`\hat{y}` is a
            tensor of predictions.

        As input to ``forward`` and ``update`` the metric accepts the following
            input:

        - ``preds`` (:class:`~torch.Tensor`): Predictions from model
        - ``target`` (:class:`~torch.Tensor`): Ground truth values

        As output of ``forward`` and ``compute`` the metric returns the
            following output:

        - ``mse_log`` (:class:`~torch.Tensor`): A tensor with the
            relative mean absolute error over the state

        Args:
            squared: If True returns MSELog value, if False returns EMSELog
                value.
            kwargs: Additional keyword arguments, see `Advanced metric settings
                <https://torchmetrics.readthedocs.io/en/stable/pages/overview.html#metric-kwargs>`_.

        Reference:
            As in e.g. From big to small: Multi-scale local planar guidance for
                monocular depth estimation
        """
        super().__init__(squared, **kwargs)

    def update(self, pred: Tensor, target: Tensor) -> None:
        """Update state with predictions and targets."""
        return super().update(pred.log(), target.log())



from torch import nn

from torch_uncertainty.layers.packed import PackedConv2d, PackedLinear

from .base import VGG, _vgg
from .configs import cfgs

__all__ = [
    "packed_vgg",
]


def packed_vgg(
    in_channels: int,
    num_classes: int,
    arch: int,
    alpha: int,
    num_estimators: int,
    gamma: int,
    norm: type[nn.Module] = nn.Identity,
    groups: int = 1,
    dropout_rate: float = 0.5,
    style: str = "imagenet",
) -> VGG:
    if arch == 11:
        config = cfgs["A"]
    elif arch == 13:  # coverage: ignore
        config = cfgs["B"]
    elif arch == 16:  # coverage: ignore
        config = cfgs["D"]
    elif arch == 19:  # coverage: ignore
        config = cfgs["E"]
    else:
        raise ValueError(f"Unknown VGG arch {arch}.")
    return _vgg(
        vgg_cfg=config,
        in_channels=in_channels,
        num_classes=num_classes,
        linear_layer=PackedLinear,
        conv2d_layer=PackedConv2d,
        norm=norm,
        groups=groups,
        dropout_rate=dropout_rate,
        style=style,
        alpha=alpha,
        num_estimators=num_estimators,
        gamma=gamma,
    )



import logging
from collections.abc import Callable
from pathlib import Path

import torch
from einops import rearrange
from torch import Tensor
from torchvision.datasets import ImageFolder
from torchvision.datasets.utils import check_integrity, download_and_extract_archive


class CUB(ImageFolder):
    base_folder = "CUB_200_2011/images"
    url = "https://data.caltech.edu/records/65de6-vp158/files/CUB_200_2011.tgz"
    filename = "CUB_200_2011.tgz"
    tgz_md5 = "97eceeb196236b17998738112f37df78"

    def __init__(
        self,
        root: str | Path,
        train: bool = True,
        transform: Callable | None = None,
        target_transform: Callable | None = None,
        return_attributes: bool = False,
        download: bool = False,
    ):
        """The Caltech-UCSD Birds-200-2011 dataset.

        Args:
            root (str): Root directory of the dataset.
            train (bool, optional): If True, creates dataset from training set, otherwise creates
                from test set. Defaults to True.
            transform (callable, optional): A function/transform that takes in an PIL image and
                returns a transformed version. E.g, transforms.RandomCrop. Defaults to None.
            target_transform (callable, optional): A function/transform that takes in the target
                and transforms it. Defaults to None.
            return_attributes (bool, optional): If True, returns the attributes instead of the images.
                Defaults to False.
            download (bool, optional): If True, downloads the dataset from the internet and puts it
                in root directory. If dataset is already downloaded, it is not downloaded again.
                Defaults to False.

        Reference:
            Wah, C. and Branson, S. and Welinder, P. and Perona, P. and Belongie, S. Caltech-UCSD
                Birds 200.
        """
        self.folder_root = Path(root)
        self.train = train
        if download:
            self._download()

        if not self._check_integrity():
            raise RuntimeError(
                "Dataset not found or corrupted. You can use download=True to download it."
            )

        super().__init__(Path(root) / "CUB_200_2011" / "images", transform, target_transform)

        training_idx = self._load_train_idx()
        self.attributes, self.uncertainties = self._load_attributes()
        self.attribute_names = self._load_attribute_names()
        self.classnames = self._load_classnames()

        self.samples = [sample for i, sample in enumerate(self.samples) if training_idx[i] == train]
        self._labels = [label for i, label in enumerate(self.targets) if training_idx[i] == train]
        self.attributes = rearrange(
            torch.masked_select(self.attributes, training_idx.unsqueeze(-1) == train),
            "(n c) -> n c",
            c=312,
        )
        self.uncertainties = rearrange(
            torch.masked_select(self.uncertainties, training_idx.unsqueeze(-1) == train),
            "(n c) -> n c",
            c=312,
        )

        if return_attributes:
            self.samples = zip(self.attributes, [sam[1] for sam in self.samples], strict=False)
            self.loader = torch.nn.Identity()

    def _load_classnames(self) -> list[str]:
        """Load the classnames of the dataset.

        Returns:
            list[str]: the list containing the names of the 200 classes.
        """
        with Path(self.folder_root / "CUB_200_2011" / "classes.txt").open("r") as f:
            return [
                line.split(" ")[1].split(".")[1].replace("\n", "").replace("_", " ") for line in f
            ]

    def _load_train_idx(self) -> Tensor:
        """Load the index of the training data to make the split.

        Returns:
            Tensor: whether the images belong to the training or test split.
        """
        with (self.folder_root / "CUB_200_2011" / "train_test_split.txt").open("r") as f:
            return torch.as_tensor([int(line.split(" ")[1]) for line in f])

    def _load_attributes(self) -> tuple[Tensor, Tensor]:
        """Load the attributes associated to each image.

        Returns:
            tuple[Tensor, Tensor]: The presence of the 312 attributes along with their uncertainty.
                The uncertainty is 0 for certain samples and 1 for non-visible attributes.
        """
        attributes, uncertainty = [], []
        with (self.folder_root / "CUB_200_2011" / "attributes" / "image_attribute_labels.txt").open(
            "r"
        ) as f:
            for line in f:
                attributes.append(int(line.split(" ")[2]))
                uncertainty.append(1 - (int(line.split(" ")[3]) - 1) / 3)
        return rearrange(torch.as_tensor(attributes), "(n c) -> n c", c=312), rearrange(
            torch.as_tensor(uncertainty), "(n c) -> n c", c=312
        )

    def _load_attribute_names(self) -> list[str]:
        """Load the names of the attributes.

        Returns:
            list[str]: The list of the names of the 312 attributes.
        """
        with (self.folder_root / "attributes.txt").open("r") as f:
            return [line.split(" ")[1].replace("\n", "").replace("_", " ") for line in f]

    def _check_integrity(self) -> bool:
        """Check the integrity of the dataset.

        Returns:
            bool: True when the md5 of the archive corresponds.
        """
        fpath = self.folder_root / self.filename
        return check_integrity(
            fpath,
            self.tgz_md5,
        )

    def _download(self):
        """Download the dataset from caltec.edu."""
        if self._check_integrity():
            logging.info("Files already downloaded and verified")
            return

        download_and_extract_archive(
            url=self.url, download_root=self.folder_root, filename=self.filename, md5=self.tgz_md5
        )



import torch
from torch import nn

from torch_uncertainty.baselines.classification import (
    ResNetBaseline,
    VGGBaseline,
    WideResNetBaseline,
)


class TestStandardBaseline:
    """Testing the ResNetBaseline baseline class."""

    def test_standard(self):
        net = ResNetBaseline(
            num_classes=10,
            in_channels=3,
            loss=nn.CrossEntropyLoss(),
            version="mc-dropout",
            dropout_rate=0.1,
            num_estimators=4,
            arch=18,
            style="cifar",
            groups=1,
        )
        net(torch.rand(1, 3, 32, 32))


class TestStandardWideBaseline:
    """Testing the WideResNetBaseline baseline class."""

    def test_standard(self):
        net = WideResNetBaseline(
            num_classes=10,
            in_channels=3,
            loss=nn.CrossEntropyLoss(),
            version="mc-dropout",
            dropout_rate=0.1,
            num_estimators=4,
            style="cifar",
            groups=1,
        )
        net(torch.rand(1, 3, 32, 32))


class TestStandardVGGBaseline:
    """Testing the VGGBaseline baseline class."""

    def test_standard(self):
        net = VGGBaseline(
            num_classes=10,
            in_channels=3,
            loss=nn.CrossEntropyLoss(),
            version="mc-dropout",
            dropout_rate=0.1,
            num_estimators=4,
            arch=11,
            groups=1,
            last_layer_dropout=True,
        )
        net(torch.rand(1, 3, 32, 32))

        net = VGGBaseline(
            num_classes=10,
            in_channels=3,
            loss=nn.CrossEntropyLoss(),
            version="mc-dropout",
            num_estimators=4,
            arch=11,
            groups=1,
            dropout_rate=0.3,
            last_layer_dropout=True,
        )
        net.eval()
        net(torch.rand(1, 3, 32, 32))



import torch
from torch import Tensor, nn
from torch.nn.modules.dropout import _DropoutNd


class MCDropout(nn.Module):
    def __init__(
        self,
        model: nn.Module,
        num_estimators: int,
        last_layer: bool,
        on_batch: bool,
    ) -> None:
        """MC Dropout wrapper for a model containing nn.Dropout modules.

        Args:
            model (nn.Module): model to wrap
            num_estimators (int): number of estimators to use during the
                evaluation
            last_layer (bool): whether to apply dropout to the last layer only.
            on_batch (bool): Perform the MC-Dropout on the batch-size.
                Otherwise in a for loop. Useful when constrained in memory.

        Warning:
            This module will work only if you apply dropout through modules
            declared in the constructor (__init__).

        Warning:
            The `last-layer` option disables the lastly initialized dropout
            during evaluation: make sure that the last dropout is either
            functional or a module of its own.
        """
        super().__init__()
        filtered_modules = list(
            filter(
                lambda m: isinstance(m, _DropoutNd),
                model.modules(),
            )
        )
        if last_layer:
            filtered_modules = filtered_modules[-1:]

        _dropout_checks(filtered_modules, num_estimators)
        self.last_layer = last_layer
        self.on_batch = on_batch
        self.core_model = model
        self.num_estimators = num_estimators
        self.filtered_modules = filtered_modules

    def train(self, mode: bool = True) -> nn.Module:
        """Override the default train method to set the training mode of
        each submodule to be the same as the module itself except for the
        selected dropout modules.

        Args:
            mode (bool, optional): whether to set the module to training
                mode. Defaults to True.
        """
        if not isinstance(mode, bool):
            raise TypeError("Training mode is expected to be boolean")
        self.training = mode
        for module in self.children():
            module.train(mode)
        for module in self.filtered_modules:
            module.train()
        return self

    def forward(
        self,
        x: Tensor,
    ) -> Tensor:
        """Forward pass of the model.

        During training, the forward pass is the same as of the core model.
        During evaluation, the forward pass is repeated `num_estimators` times
        either on the batch size or in a for loop depending on
        :attr:`last_layer`.

        Args:
            x (Tensor): input tensor of shape (B, ...)

        Returns:
            Tensor: output tensor of shape (:attr:`num_estimators` * B, ...)
        """
        if self.training:
            return self.core_model(x)
        if self.on_batch:
            x = x.repeat(self.num_estimators, 1, 1, 1)
            return self.core_model(x)
        # Else, for loop
        return torch.cat([self.core_model(x) for _ in range(self.num_estimators)], dim=0)


def mc_dropout(
    model: nn.Module,
    num_estimators: int,
    last_layer: bool = False,
    on_batch: bool = True,
) -> MCDropout:
    """MC Dropout wrapper for a model.

    Args:
        model (nn.Module): model to wrap
        num_estimators (int): number of estimators to use
        last_layer (bool, optional): whether to apply dropout to the last
            layer only. Defaults to False.
        on_batch (bool): Increase the batch_size to perform MC-Dropout.
            Otherwise in a for loop to reduce memory footprint. Defaults
            to true.
    """
    return MCDropout(
        model=model,
        num_estimators=num_estimators,
        last_layer=last_layer,
        on_batch=on_batch,
    )


def _dropout_checks(filtered_modules: list[nn.Module], num_estimators: int) -> None:
    if not filtered_modules:
        raise ValueError(
            "No dropout module found in the model. "
            "Please use `nn.Dropout`-like modules to apply dropout."
        )
    # Check that at least one module has > 0.0 dropout rate
    if not any(mod.p > 0.0 for mod in filtered_modules):
        raise ValueError("At least one dropout module must have a dropout rate > 0.0.")
    if num_estimators <= 0:
        raise ValueError("`num_estimators` must be strictly positive to use MC Dropout.")



import copy

from torch import Tensor, nn


class EMA(nn.Module):
    def __init__(
        self,
        model: nn.Module,
        momentum: float,
    ) -> None:
        """Exponential Moving Average.

        Args:
            model (nn.Module): The model to train and ensemble.
            momentum (float): The momentum of the moving average.
        """
        super().__init__()
        _ema_checks(momentum)
        self.core_model = model
        self.ema_model = copy.deepcopy(model)
        self.momentum = momentum
        self.remainder = 1 - momentum

    def update_wrapper(self, epoch: int | None = None) -> None:
        """Update the EMA model.

        Args:
            epoch (int): The current epoch. For API consistency.
        """
        for ema_param, param in zip(
            self.ema_model.parameters(),
            self.core_model.parameters(),
            strict=False,
        ):
            ema_param.data = ema_param.data * self.momentum + param.data * self.remainder

    def eval_forward(self, x: Tensor) -> Tensor:
        return self.ema_model.forward(x)

    def forward(self, x: Tensor) -> Tensor:
        if self.training:
            return self.core_model.forward(x)
        return self.eval_forward(x)


def _ema_checks(momentum: float) -> None:
    if momentum < 0.0 or momentum >= 1.0:
        raise ValueError(f"`momentum` must be in the range [0, 1). Got {momentum}.")



import pytest
import torch

from torch_uncertainty.metrics.classification import FPR95, FPRx


class TestFPR95:
    """Testing the Entropy metric class."""

    def test_compute_zero(self):
        metric = FPR95(pos_label=1)
        metric.update(torch.as_tensor([1] * 99 + [0.99]), torch.as_tensor([1] * 99 + [0]))
        res = metric.compute()
        assert res == 0

    def test_compute_half(self):
        metric = FPR95(pos_label=1)
        metric.update(
            torch.as_tensor([0.9] * 100 + [0.95] * 50 + [0.85] * 50),
            torch.as_tensor([1] * 100 + [0] * 100),
        )
        res = metric.compute()
        assert res == 0.5

    def test_compute_one(self):
        metric = FPR95(pos_label=1)
        metric.update(torch.as_tensor([0.99] * 99 + [1]), torch.as_tensor([1] * 99 + [0]))
        res = metric.compute()
        assert res == 1

    def test_compute_nan(self):
        metric = FPR95(pos_label=1)
        metric.update(torch.as_tensor([0.1] * 50 + [0.4] * 50), torch.as_tensor([0] * 100))
        res = metric.compute()
        assert torch.isnan(res).all()

    def test_error(self):
        with pytest.raises(ValueError):
            FPRx(recall_level=1.2, pos_label=1)



"""
Train a Bayesian Neural Network in Three Minutes
================================================

In this tutorial, we will train a variational inference Bayesian Neural Network (BNN) LeNet classifier on the MNIST dataset.

Foreword on Bayesian Neural Networks
------------------------------------

Bayesian Neural Networks (BNNs) are a class of neural networks that estimate the uncertainty on their predictions via uncertainty
on their weights. This is achieved by considering the weights of the neural network as random variables, and by learning their
posterior distribution. This is in contrast to standard neural networks, which only learn a single set of weights, which can be
seen as Dirac distributions on the weights.

For more information on Bayesian Neural Networks, we refer the reader to the following resources:

- Weight Uncertainty in Neural Networks `ICML2015 <https://arxiv.org/pdf/1505.05424.pdf>`_
- Hands-on Bayesian Neural Networks - a Tutorial for Deep Learning Users `IEEE Computational Intelligence Magazine <https://arxiv.org/pdf/2007.06823.pdf>`_

Training a Bayesian LeNet using TorchUncertainty models and Lightning
---------------------------------------------------------------------

In this part, we train a Bayesian LeNet, based on the model and routines already implemented in TU.

1. Loading the utilities
~~~~~~~~~~~~~~~~~~~~~~~~

To train a BNN using TorchUncertainty, we have to load the following modules:

- our TUTrainer
- the model: bayesian_lenet, which lies in the torch_uncertainty.model
- the classification training routine from torch_uncertainty.routines
- the Bayesian objective: the ELBOLoss, which lies in the torch_uncertainty.losses file
- the datamodule that handles dataloaders: MNISTDataModule from torch_uncertainty.datamodules

We will also need to define an optimizer using torch.optim and Pytorch's
neural network utils from torch.nn.
"""
# %%
from pathlib import Path

from torch import nn, optim

from torch_uncertainty import TUTrainer
from torch_uncertainty.datamodules import MNISTDataModule
from torch_uncertainty.losses import ELBOLoss
from torch_uncertainty.models.lenet import bayesian_lenet
from torch_uncertainty.routines import ClassificationRoutine

# %%
# 2. The Optimization Recipe
# ~~~~~~~~~~~~~~~~~~~~~~~~~~
# We will use the Adam optimizer with the default learning rate of 0.001.


def optim_lenet(model: nn.Module):
    optimizer = optim.Adam(
        model.parameters(),
        lr=1e-3,
    )
    return optimizer


# %%
# 3. Creating the necessary variables
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#
# In the following, we instantiate our trainer, define the root of the datasets and the logs.
# We also create the datamodule that handles the MNIST dataset, dataloaders and transforms.
# Please note that the datamodules can also handle OOD detection by setting the eval_ood
# parameter to True. Finally, we create the model using the blueprint from torch_uncertainty.models.

trainer = TUTrainer(accelerator="cpu", enable_progress_bar=False, max_epochs=1)

# datamodule
root = Path("data")
datamodule = MNISTDataModule(root=root, batch_size=128, eval_ood=False)

# model
model = bayesian_lenet(datamodule.num_channels, datamodule.num_classes)

# %%
# 4. The Loss and the Training Routine
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Then, we just have to define the loss to be used during training. To do this,
# we redefine the default parameters from the ELBO loss using the partial
# function from functools. We use the hyperparameters proposed in the blitz
# library. As we are train a classification model, we use the CrossEntropyLoss
# as the likelihood.
# We then define the training routine using the classification training routine
# from torch_uncertainty.classification. We provide the model, the ELBO
# loss and the optimizer to the routine.

loss = ELBOLoss(
    model=model,
    inner_loss=nn.CrossEntropyLoss(),
    kl_weight=1 / 10000,
    num_samples=3,
)

routine = ClassificationRoutine(
    model=model,
    num_classes=datamodule.num_classes,
    loss=loss,
    optim_recipe=optim_lenet(model),
    is_ensemble=True
)

# %%
# 5. Gathering Everything and Training the Model
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#
# Now that we have prepared all of this, we just have to gather everything in
# the main function and to train the model using our wrapper of Lightning Trainer.
# Specifically, it needs the routine, that includes the model as well as the
# training/eval logic and the datamodule
# The dataset will be downloaded automatically in the root/data folder, and the
# logs will be saved in the root/logs folder.

trainer.fit(model=routine, datamodule=datamodule)
trainer.test(model=routine, datamodule=datamodule)

# %%
# 6. Testing the Model
# ~~~~~~~~~~~~~~~~~~~~
#
# Now that the model is trained, let's test it on MNIST.
# Please note that we apply a reshape to the logits to determine the dimension corresponding to the ensemble
# and to the batch. As for TorchUncertainty 0.2.0, the ensemble dimension is merged with the batch dimension
# in this order (num_estimator x batch, classes).
import matplotlib.pyplot as plt
import numpy as np
import torch
import torchvision


def imshow(img):
    npimg = img.numpy()
    plt.imshow(np.transpose(npimg, (1, 2, 0)))
    plt.axis("off")
    plt.tight_layout()
    plt.show()


dataiter = iter(datamodule.val_dataloader())
images, labels = next(dataiter)

# print images
imshow(torchvision.utils.make_grid(images[:4, ...]))
print("Ground truth: ", " ".join(f"{labels[j]}" for j in range(4)))

# Put the model in eval mode to use several samples
model = model.eval()
logits = model(images).reshape(16, 128, 10) # num_estimators, batch_size, num_classes

# We apply the softmax on the classes and average over the estimators
probs = torch.nn.functional.softmax(logits, dim=-1)
avg_probs = probs.mean(dim=0)
var_probs = probs.std(dim=0)

_, predicted = torch.max(avg_probs, 1)

print("Predicted digits: ", " ".join(f"{predicted[j]}" for j in range(4)))
print("Std. dev. of the scores over the posterior samples", " ".join(f"{var_probs[j][predicted[j]]:.3}" for j in range(4)))
# %%
# Here, we show the variance of the top prediction. This is a non-standard but intuitive way to show the diversity of the predictions
# of the ensemble. Ideally, the variance should be high when the average top prediction is incorrect.
#
# References
# ----------
#
# - **LeNet & MNIST:** LeCun, Y., Bottou, L., Bengio, Y., & Haffner, P. (1998). Gradient-based learning applied to document recognition. `Proceedings of the IEEE <http://vision.stanford.edu/cs598_spring07/papers/Lecun98.pdf>`_.
# - **Bayesian Neural Networks:** Blundell, C., Cornebise, J., Kavukcuoglu, K., & Wierstra, D. (2015). Weight Uncertainty in Neural Networks. `ICML 2015 <https://arxiv.org/pdf/1505.05424.pdf>`_.
# - **The Adam optimizer:** Kingma, D. P., & Ba, J. (2014). "Adam: A method for stochastic optimization." `ICLR 2015 <https://arxiv.org/pdf/1412.6980.pdf>`_.
# - **The Blitz** `library <https://github.com/piEsposito/blitz-bayesian-deep-learning>`_ (for the hyperparameters).



from .base import ImageNetVariation


class ImageNetO(ImageNetVariation):
    url = "https://people.eecs.berkeley.edu/~hendrycks/imagenet-o.tar"
    filename = "imagenet-o.tar"
    tgz_md5 = "86bd7a50c1c4074fb18fc5f219d6d50b"
    dataset_name = "imagenet-o"



import pytest
from torch import nn
from torchvision.datasets import MNIST

from tests._dummies.dataset import DummyClassificationDataset
from torch_uncertainty.datamodules import MNISTDataModule
from torch_uncertainty.transforms import Cutout


class TestMNISTDataModule:
    """Testing the MNISTDataModule datamodule class."""

    def test_mnist_cutout(self):
        dm = MNISTDataModule(
            root="./data/",
            batch_size=128,
            cutout=16,
            val_split=0.1,
            eval_ood=True,
        )

        assert dm.dataset == MNIST
        assert isinstance(dm.train_transform.transforms[2], Cutout)

        dm = MNISTDataModule(
            root="./data/",
            batch_size=128,
            ood_ds="notMNIST",
            cutout=0,
            val_split=0,
            basic_augment=False,
        )
        assert isinstance(dm.train_transform.transforms[2], nn.Identity)

        with pytest.raises(ValueError):
            MNISTDataModule(root="./data/", batch_size=128, ood_ds="other")

        dm.dataset = DummyClassificationDataset
        dm.ood_dataset = DummyClassificationDataset
        dm.shift_dataset = DummyClassificationDataset
        dm.setup("fit")
        dm.setup("test")
        dm.train_dataloader()
        dm.val_dataloader()
        dm.test_dataloader()

        with pytest.raises(ValueError):
            dm.setup("other")

        dm.eval_ood = True
        dm.eval_shift = True
        dm.ood_transform = dm.test_transform
        dm.val_split = 0.1
        dm.prepare_data()
        dm.setup()
        dm.test_dataloader()



from torch import Tensor
from torchmetrics.classification.stat_scores import MulticlassStatScores
from torchmetrics.utilities.compute import _safe_divide


class MeanIntersectionOverUnion(MulticlassStatScores):
    is_differentiable = False
    higher_is_better = True
    full_state_update = False

    def __init__(
        self,
        num_classes: int,
        top_k: int = 1,
        ignore_index: int | None = None,
        validate_args: bool = True,
        **kwargs,
    ) -> None:
        r"""Computes Mean Intersection over Union (IoU) score.

        Args:
            num_classes (int): Integer specifying the number of classes.
            top_k (int, optional): Number of highest probability or logit score predictions
                considered to find the correct label. Only works when ``preds`` contain
                probabilities/logits. Defaults to ``1``.
            ignore_index (int | None, optional): Specifies a target value that is ignored and does
                not contribute to the metric calculation. Defaults to ``None``.
            validate_args (bool, optional): Bool indicating if input arguments and tensors should
                be validated for correctness. Set to ``False`` for faster computations. Defaults to
                ``True``.
            **kwargs: kwargs: Additional keyword arguments, see
                `Advanced metric settings <https://lightning.ai/docs/torchmetrics/stable/pages/overview.html#metric-kwargs>`_
                for more info.

        Shape:
            As input to ``forward`` and ``update`` the metric accepts the following input:

            - ``preds`` (:class:`~torch.Tensor`): An int tensor of shape ``(B, ...)`` or float tensor of shape ``(B, C, ..)``.
              If preds is a floating point we apply ``torch.argmax`` along the ``C`` dimension to automatically convert
              probabilities/logits into an int tensor.
            - ``target`` (:class:`~torch.Tensor`): An int tensor of shape ``(B, ...)``.

            As output to ``forward`` and ``compute`` the metric returns the following output:

            - ``mean_iou`` (:class:`~torch.Tensor`): The computed Mean Intersection over Union (IoU) score.
              A tensor containing a single float value.
        """
        super().__init__(
            num_classes,
            top_k,
            "macro",
            "global",
            ignore_index,
            validate_args,
            **kwargs,
        )

    def compute(self) -> Tensor:
        """Compute the Means Intersection over Union (MIoU) based on saved inputs."""
        tp, fp, _, fn = self._final_state()

        return _safe_divide(tp, tp + fp + fn, zero_division=float("nan")).nanmean()



import torch
from torch import Tensor
from torchmetrics import Metric
from torchmetrics.utilities import rank_zero_warn
from torchmetrics.utilities.data import dim_zero_cat


class FPRx(Metric):
    is_differentiable = False
    higher_is_better = False
    full_state_update = False

    conf: list[Tensor]
    targets: list[Tensor]

    def __init__(self, recall_level: float, pos_label: int, **kwargs) -> None:
        """The False Positive Rate at x% Recall metric.

        Args:
            recall_level (float): The recall level at which to compute the FPR.
            pos_label (int): The positive label.
            kwargs: Additional arguments to pass to the metric class.

        Reference:
            Improved from https://github.com/hendrycks/anomaly-seg and
            translated to torch.
        """
        super().__init__(**kwargs)

        if recall_level < 0 or recall_level > 1:
            raise ValueError(f"Recall level must be between 0 and 1. Got {recall_level}.")
        self.recall_level = recall_level
        self.pos_label = pos_label
        self.add_state("conf", [], dist_reduce_fx="cat")
        self.add_state("targets", [], dist_reduce_fx="cat")

        rank_zero_warn(
            f"Metric `FPR{int(recall_level * 100)}` will save all targets and predictions"
            " in buffer. For large datasets this may lead to large memory"
            " footprint."
        )

    def update(self, conf: Tensor, target: Tensor) -> None:
        """Update the metric state.

        Args:
            conf (Tensor): The confidence scores.
            target (Tensor): The target labels, 0 if ID, 1 if OOD.
        """
        self.conf.append(conf)
        self.targets.append(target)

    def compute(self) -> Tensor:
        """Compute the False Positive Rate at x% Recall.

        Returns:
            Tensor: The value of the FPRx.
        """
        conf = dim_zero_cat(self.conf)
        targets = dim_zero_cat(self.targets)

        # map examples and labels to OOD first
        indx = torch.argsort(targets, descending=True)
        examples = conf[indx]
        labels = torch.zeros_like(targets, dtype=torch.bool, device=self.device)
        labels[: torch.count_nonzero(targets)] = True

        # sort examples and labels by decreasing confidence
        desc_scores_indx = torch.argsort(examples, descending=True)
        examples = examples[desc_scores_indx]
        labels = labels[desc_scores_indx]

        # Get the indices of the distinct values
        distinct_value_indices = torch.where(torch.diff(examples))[0]
        threshold_idxs = torch.cat(
            [
                distinct_value_indices,
                torch.tensor([labels.shape[0] - 1], dtype=torch.long, device=self.device),
            ]
        )

        # accumulate the true positives with decreasing threshold
        true_pos = torch.cumsum(labels, dim=0)[threshold_idxs]
        false_pos = 1 + threshold_idxs - true_pos  # add one because of zero-based indexing

        # check that there is at least one OOD example
        if true_pos[-1] == 0:
            return torch.tensor([torch.nan], device=self.device)

        recall = true_pos / true_pos[-1]

        last_ind = torch.searchsorted(true_pos, true_pos[-1])
        recall = torch.cat(
            [
                recall[: last_ind + 1].flip(0),
                torch.tensor([1.0], device=self.device),
            ]
        )
        false_pos = torch.cat(
            [
                false_pos[: last_ind + 1].flip(0),
                torch.tensor([0.0], device=self.device),
            ]
        )
        cutoff = torch.argmin(torch.abs(recall - self.recall_level))
        return false_pos[cutoff] / (~labels).sum()


class FPR95(FPRx):
    def __init__(self, pos_label: int, **kwargs) -> None:
        """The False Positive Rate at 95% Recall metric.

        Args:
            pos_label (int): The positive label.
            kwargs: Additional arguments to pass to the metric class.
        """
        super().__init__(recall_level=0.95, pos_label=pos_label, **kwargs)



import pytest
import torch
from requests.exceptions import HTTPError

from torch_uncertainty.transforms.corruption import (
    Brightness,
    Contrast,
    DefocusBlur,
    Elastic,
    Fog,
    Frost,
    GaussianBlur,
    GaussianNoise,
    GlassBlur,
    ImpulseNoise,
    JPEGCompression,
    MotionBlur,
    Pixelate,
    Saturation,
    ShotNoise,
    Snow,
    SpeckleNoise,
    ZoomBlur,
)


class TestCorruptions:
    """Testing the Corruptions transform."""

    def test_gaussian_noise(self):
        with pytest.raises(ValueError):
            _ = GaussianNoise(-1)
        with pytest.raises(TypeError):
            _ = GaussianNoise(0.1)
        inputs = torch.rand(3, 32, 32)
        transform = GaussianNoise(1)
        transform(inputs)
        transform = GaussianNoise(0)
        transform(inputs)
        print(transform)

    def test_shot_noise(self):
        inputs = torch.rand(3, 32, 32)
        transform = ShotNoise(1)
        transform(inputs)
        transform = ShotNoise(0)
        transform(inputs)

    def test_impulse_noise(self):
        inputs = torch.rand(3, 32, 32)
        transform = ImpulseNoise(1)
        transform(inputs)
        transform = ImpulseNoise(0)
        transform(inputs)

    def test_speckle_noise(self):
        inputs = torch.rand(3, 32, 32)
        transform = SpeckleNoise(1)
        transform(inputs)
        transform = SpeckleNoise(0)
        transform(inputs)

    def test_gaussian_blur(self):
        inputs = torch.rand(3, 32, 32)
        transform = GaussianBlur(1)
        transform(inputs)
        transform = GaussianBlur(0)
        transform(inputs)

    def test_glass_blur(self):
        inputs = torch.rand(3, 32, 32)
        transform = GlassBlur(1)
        transform(inputs)
        transform = GlassBlur(0)
        transform(inputs)

    def test_defocus_blur(self):
        inputs = torch.rand(3, 32, 32)
        transform = DefocusBlur(1)
        transform(inputs)
        transform = DefocusBlur(0)
        transform(inputs)

    def test_motion_blur(self):
        inputs = torch.rand(3, 32, 32)
        transform = MotionBlur(1)
        transform(inputs)
        transform = MotionBlur(0)
        transform(inputs)

        inputs = torch.rand(1, 3, 32, 32)
        transform = MotionBlur(1)
        transform(inputs)

    def test_zoom_blur(self):
        inputs = torch.rand(3, 32, 32)
        transform = ZoomBlur(1)
        transform(inputs)
        transform = ZoomBlur(0)
        transform(inputs)

    def test_jpeg_compression(self):
        inputs = torch.rand(3, 32, 32)
        transform = JPEGCompression(1)
        transform(inputs)
        transform = JPEGCompression(0)
        transform(inputs)

    def test_pixelate(self):
        inputs = torch.rand(3, 32, 32)
        transform = Pixelate(1)
        transform(inputs)
        transform = Pixelate(0)
        transform(inputs)

    def test_frost(self):
        try:
            Frost(1)
            frost_ok = True
        except HTTPError:
            frost_ok = False
        if frost_ok:
            inputs = torch.rand(3, 32, 32)
            transform = Frost(1)
            transform(inputs)
            transform = Frost(0)
            transform(inputs)

    def test_snow(self):
        inputs = torch.rand(3, 32, 32)
        transform = Snow(1)
        transform(inputs)
        transform = Snow(0)
        transform(inputs)

    def test_fog(self):
        inputs = torch.rand(3, 32, 32)
        transform = Fog(1, size=32)
        transform(inputs)

        with pytest.raises(ValueError, match="Image must be square. Got "):
            transform(torch.rand(3, 32, 12))

        transform = Fog(0, size=32)
        transform(inputs)

        with pytest.raises(ValueError, match="Size must be a power of 2. Got "):
            _ = Fog(1, size=15)

    def test_brightness(self):
        inputs = torch.rand(3, 32, 32)
        transform = Brightness(1)
        transform(inputs)
        transform = Brightness(0)
        transform(inputs)

    def test_contrast(self):
        inputs = torch.rand(3, 32, 32)
        transform = Contrast(1)
        transform(inputs)
        transform = Contrast(0)
        transform(inputs)

    def test_elastic(self):
        inputs = torch.rand(3, 32, 32)
        transform = Elastic(1)
        transform(inputs)
        transform = Elastic(0)
        transform(inputs)

    def test_saturation(self):
        inputs = torch.rand(3, 32, 32)
        transform = Saturation(1)
        transform(inputs)
        transform = Saturation(0)
        transform(inputs)



import math

import pytest
import torch

from torch_uncertainty.metrics import Entropy


@pytest.fixture()
def vec2d_max() -> torch.Tensor:
    vec = torch.as_tensor([0.5, 0.5])
    return vec.unsqueeze(0)


@pytest.fixture()
def vec2d_min() -> torch.Tensor:
    vec = torch.as_tensor([0.0, 1.0])
    return vec.unsqueeze(0)


@pytest.fixture()
def vec3d() -> torch.Tensor:
    """Return a torch tensor with a mean entropy of 0 and an entropy of
    the mean of ln(2) to test the `ensemble` parameter of `Entropy`.
    """
    vec = torch.as_tensor([[0.0, 1.0], [1.0, 0.0]])
    return vec.unsqueeze(0)


class TestEntropy:
    """Testing the Entropy metric class."""

    def test_compute(self, vec2d_min: torch.Tensor):
        metric = Entropy()
        metric.update(vec2d_min)
        res = metric.compute()
        assert res == 0

    def test_compute_max(self, vec2d_max: torch.Tensor):
        metric = Entropy(reduction="sum")
        metric.update(vec2d_max)
        res = metric.compute()
        assert res == math.log(2)

    def test_multiple_compute_sum(self, vec2d_min: torch.Tensor, vec2d_max: torch.Tensor):
        metric = Entropy(reduction="sum")
        metric.update(vec2d_min)
        metric.update(vec2d_max)
        res = metric.compute()
        assert res == math.log(2)

    def test_multiple_compute_mean(self, vec2d_min: torch.Tensor, vec2d_max: torch.Tensor):
        metric = Entropy(reduction="mean")
        metric.update(vec2d_min)
        metric.update(vec2d_max)
        res = metric.compute()
        assert res == math.log(2) / 2

    def test_multiple_compute_none(self, vec2d_min: torch.Tensor, vec2d_max: torch.Tensor):
        metric = Entropy(reduction=None)
        metric.update(vec2d_min)
        metric.update(vec2d_max)
        res = metric.compute()
        assert all(res == torch.as_tensor([0, math.log(2)]))

    def test_compute_3d(self, vec3d: torch.Tensor):
        metric = Entropy(reduction="mean")
        metric.update(vec3d)
        res = metric.compute()
        assert res == 0

    def test_compute_3d_to_2d(self, vec3d: torch.Tensor):
        metric = Entropy(reduction="mean")
        vec3d = vec3d.mean(1)
        metric.update(vec3d)
        res = metric.compute()
        assert res == math.log(2)

    def test_bad_argument(self):
        with pytest.raises(ValueError):
            _ = Entropy("geometric_mean")



# ruff: noqa: F401
from .abstract import TUDataModule
from .classification import (
    BankMarketingDataModule,
    CIFAR10DataModule,
    CIFAR100DataModule,
    DOTA2GamesDataModule,
    HTRU2DataModule,
    ImageNetDataModule,
    MNISTDataModule,
    OnlineShoppersDataModule,
    SpamBaseDataModule,
    TinyImageNetDataModule,
)
from .segmentation import CamVidDataModule, CityscapesDataModule
from .uci_regression import UCIRegressionDataModule



import pytest
import torch

from torch_uncertainty.layers.masksembles import MaskedConv2d, MaskedLinear


@pytest.fixture()
def feat_input_odd() -> torch.Tensor:
    return torch.rand((5, 10))


@pytest.fixture()
def feat_input_even() -> torch.Tensor:
    return torch.rand((8, 10))


@pytest.fixture()
def img_input_odd() -> torch.Tensor:
    return torch.rand((5, 10, 3, 3))


@pytest.fixture()
def img_input_even() -> torch.Tensor:
    return torch.rand((8, 10, 3, 3))


class TestMaskedLinear:
    """Testing the MaskedLinear layer class."""

    def test_linear_one_estimator(self, feat_input_odd: torch.Tensor):
        layer = MaskedLinear(10, 2, num_estimators=1, scale=2)
        out = layer(feat_input_odd)
        assert out.shape == torch.Size([5, 2])

    def test_linear_two_estimators_odd(self, feat_input_odd: torch.Tensor):
        layer = MaskedLinear(10, 2, num_estimators=2, scale=2)
        with pytest.raises(RuntimeError):
            _ = layer(feat_input_odd)

    def test_linear_two_estimators_even(self, feat_input_even: torch.Tensor):
        layer = MaskedLinear(10, 2, num_estimators=2, scale=2)
        out = layer(feat_input_even)
        assert out.shape == torch.Size([8, 2])

    def test_linear_errors(self):
        with pytest.raises(ValueError):
            _ = MaskedLinear(8, 2, num_estimators=1, scale=2)

        with pytest.raises(ValueError):
            _ = MaskedLinear(8, 2, num_estimators=1, scale=None)

        with pytest.raises(ValueError):
            _ = MaskedLinear(10, 2, num_estimators=1, scale=7)

        with pytest.raises(ValueError):
            _ = MaskedLinear(10, 2, num_estimators=1, scale=0)


class TestMaskedConv2d:
    """Testing the MaskedConv2d layer class."""

    def test_conv_one_estimator(self, img_input_odd: torch.Tensor):
        layer = MaskedConv2d(10, 2, num_estimators=1, kernel_size=1, scale=2)
        out = layer(img_input_odd)
        assert out.shape == torch.Size([5, 2, 3, 3])

    def test_conv_two_estimators_odd(self, img_input_odd: torch.Tensor):
        layer = MaskedConv2d(10, 2, num_estimators=2, kernel_size=1, scale=2)
        with pytest.raises(RuntimeError):
            _ = layer(img_input_odd)

    def test_conv_two_estimators_even(self, img_input_even: torch.Tensor):
        layer = MaskedConv2d(10, 2, num_estimators=2, kernel_size=1, scale=2)
        out = layer(img_input_even)
        assert out.shape == torch.Size([8, 2, 3, 3])

    def test_conv_error(self):
        with pytest.raises(ValueError):
            MaskedConv2d(10, 2, num_estimators=2, kernel_size=1, scale=None)

        with pytest.raises(ValueError):
            MaskedConv2d(10, 2, num_estimators=2, kernel_size=1, scale=0)



from collections.abc import Callable
from pathlib import Path
from typing import Literal

import torch
import torch.nn.functional as F
from einops import rearrange
from lightning.pytorch import LightningModule
from lightning.pytorch.loggers import Logger
from lightning.pytorch.utilities.types import STEP_OUTPUT
from timm.data import Mixup as timm_Mixup
from torch import Tensor, nn
from torch.optim import Optimizer
from torchmetrics import Accuracy, MetricCollection
from torchmetrics.classification import (
    BinaryAUROC,
    BinaryAveragePrecision,
)

from torch_uncertainty.layers import Identity
from torch_uncertainty.losses import DECLoss, ELBOLoss
from torch_uncertainty.metrics import (
    AUGRC,
    AURC,
    FPR95,
    BrierScore,
    CalibrationError,
    CategoricalNLL,
    CovAt5Risk,
    Disagreement,
    Entropy,
    GroupingLoss,
    MutualInformation,
    RiskAt80Cov,
    VariationRatio,
)
from torch_uncertainty.models import (
    EPOCH_UPDATE_MODEL,
    STEP_UPDATE_MODEL,
)
from torch_uncertainty.post_processing import LaplaceApprox, PostProcessing
from torch_uncertainty.transforms import (
    Mixup,
    MixupIO,
    RegMixup,
    RepeatTarget,
    WarpingMixup,
)
from torch_uncertainty.utils import csv_writer, plot_hist

MIXUP_PARAMS = {
    "mixtype": "erm",
    "mixmode": "elem",
    "dist_sim": "emb",
    "kernel_tau_max": 1.0,
    "kernel_tau_std": 0.5,
    "mixup_alpha": 0,
    "cutmix_alpha": 0,
}


class ClassificationRoutine(LightningModule):
    def __init__(
        self,
        model: nn.Module,
        num_classes: int,
        loss: nn.Module,
        is_ensemble: bool = False,
        format_batch_fn: nn.Module | None = None,
        optim_recipe: dict | Optimizer | None = None,
        mixup_params: dict | None = None,
        eval_ood: bool = False,
        eval_shift: bool = False,
        eval_grouping_loss: bool = False,
        ood_criterion: Literal["msp", "logit", "energy", "entropy", "mi", "vr"] = "msp",
        post_processing: PostProcessing | None = None,
        calibration_set: Literal["val", "test"] = "val",
        num_calibration_bins: int = 15,
        log_plots: bool = False,
        save_in_csv: bool = False,
    ) -> None:
        r"""Routine for training & testing on **classification** tasks.

        Args:
            model (torch.nn.Module): Model to train.
            num_classes (int): Number of classes.
            loss (torch.nn.Module): Loss function to optimize the :attr:`model`.
            is_ensemble (bool, optional): Indicates whether the model is an
                ensemble at test time or not. Defaults to ``False``.
            format_batch_fn (torch.nn.Module, optional): Function to format the batch.
                Defaults to :class:`torch.nn.Identity()`.
            optim_recipe (dict or torch.optim.Optimizer, optional): The optimizer and
                optionally the scheduler to use. Defaults to ``None``.
            mixup_params (dict, optional): Mixup parameters. Can include mixup type,
                mixup mode, distance similarity, kernel tau max, kernel tau std,
                mixup alpha, and cutmix alpha. If None, no mixup augmentations.
                Defaults to ``None``.
            eval_ood (bool, optional): Indicates whether to evaluate the OOD
                detection performance. Defaults to ``False``.
            eval_shift (bool, optional): Indicates whether to evaluate the Distribution
                shift performance. Defaults to ``False``.
            eval_grouping_loss (bool, optional): Indicates whether to evaluate the
                grouping loss or not. Defaults to ``False``.
            ood_criterion (str, optional): OOD criterion. Available options are
                - ``"msp"`` (default): Maximum softmax probability.
                - ``"logit"``: Maximum logit.
                - ``"energy"``: Logsumexp of the mean logits.
                - ``"entropy"``: Entropy of the mean prediction.
                - ``"mi"``: Mutual information of the ensemble.
                - ``"vr"``: Variation ratio of the ensemble.
            post_processing (PostProcessing, optional): Post-processing method
                to train on the calibration set. No post-processing if None.
                Defaults to ``None``.
            calibration_set (str, optional): The post-hoc calibration dataset to
                use for the post-processing method. Defaults to ``val``.
            num_calibration_bins (int, optional): Number of bins to compute calibration
                metrics. Defaults to ``15``.
            log_plots (bool, optional): Indicates whether to log plots from
                metrics. Defaults to ``False``.
            save_in_csv(bool, optional): Save the results in csv. Defaults to
                ``False``.

        Warning:
            You must define :attr:`optim_recipe` if you do not use the Lightning CLI.

        Warning:
            When using an ensemble model, you must:
            1. Set :attr:`is_ensemble` to ``True``.
            2. Set :attr:`format_batch_fn` to :class:`torch_uncertainty.transforms.RepeatTarget(num_repeats=num_estimators)`.
            3. Ensure that the model's forward pass outputs a tensor of shape :math:`(M \times B, C)`,
            where :math:`M` is the number of estimators, :math:`B` is the batch size, :math:`C` is the number of classes.

            For automated batch handling, consider using the available model wrappers in `torch_uncertainty.models.wrappers`.

        Note:
            If :attr:`eval_ood` is ``True``, we perform a binary classification and update the
            OOD-related metrics twice:
            - once during the test on ID values where the given binary label is 0 (for ID)
            - once during the test on OOD values where the given binary label is 1 (for OOD)

        Note:
            :attr:`optim_recipe` can be anything that can be returned by
            :meth:`LightningModule.configure_optimizers()`. Find more details
            `here <https://lightning.ai/docs/pytorch/stable/common/lightning_module.html#configure-optimizers>`_.
        """
        super().__init__()
        _classification_routine_checks(
            model=model,
            num_classes=num_classes,
            is_ensemble=is_ensemble,
            ood_criterion=ood_criterion,
            eval_grouping_loss=eval_grouping_loss,
            num_calibration_bins=num_calibration_bins,
            mixup_params=mixup_params,
            post_processing=post_processing,
            format_batch_fn=format_batch_fn,
        )

        if format_batch_fn is None:
            format_batch_fn = nn.Identity()

        self.num_classes = num_classes
        self.eval_ood = eval_ood
        self.eval_shift = eval_shift
        self.eval_grouping_loss = eval_grouping_loss
        self.ood_criterion = ood_criterion
        self.log_plots = log_plots
        self.save_in_csv = save_in_csv
        self.calibration_set = calibration_set
        self.binary_cls = num_classes == 1
        self.needs_epoch_update = isinstance(model, EPOCH_UPDATE_MODEL)
        self.needs_step_update = isinstance(model, STEP_UPDATE_MODEL)
        self.num_calibration_bins = num_calibration_bins
        self.model = model
        self.loss = loss
        self.format_batch_fn = format_batch_fn
        self.optim_recipe = optim_recipe
        self.is_ensemble = is_ensemble

        self.post_processing = post_processing
        if self.post_processing is not None:
            self.post_processing.set_model(self.model)

        self._init_metrics()
        self.mixup = self._init_mixup(mixup_params)

        self.is_elbo = isinstance(self.loss, ELBOLoss)
        if self.is_elbo:
            self.loss.set_model(self.model)
        self.is_dec = isinstance(self.loss, DECLoss)

        self.id_logit_storage = None
        self.ood_logit_storage = None

    def _init_metrics(self) -> None:
        """Initialize the metrics depending on the exact task."""
        task = "binary" if self.binary_cls else "multiclass"

        metrics_dict = {
            "cls/Acc": Accuracy(task=task, num_classes=self.num_classes),
            "cls/Brier": BrierScore(num_classes=self.num_classes),
            "cls/NLL": CategoricalNLL(),
            "cal/ECE": CalibrationError(
                task=task,
                num_bins=self.num_calibration_bins,
                num_classes=self.num_classes,
            ),
            "cal/aECE": CalibrationError(
                task=task,
                adaptive=True,
                num_bins=self.num_calibration_bins,
                num_classes=self.num_classes,
            ),
            "sc/AURC": AURC(),
            "sc/AUGRC": AUGRC(),
            "sc/Cov@5Risk": CovAt5Risk(),
            "sc/Risk@80Cov": RiskAt80Cov(),
        }
        groups = [
            ["cls/Acc"],
            ["cls/Brier"],
            ["cls/NLL"],
            ["cal/ECE", "cal/aECE"],
            ["sc/AURC", "sc/AUGRC", "sc/Cov@5Risk", "sc/Risk@80Cov"],
        ]

        if self.binary_cls:
            metrics_dict |= {
                "cls/AUROC": BinaryAUROC(),
                "cls/AUPR": BinaryAveragePrecision(),
                "cls/FRP95": FPR95(pos_label=1),
            }
            groups.extend([["cls/AUROC", "cls/AUPR"], ["cls/FRP95"]])

        cls_metrics = MetricCollection(metrics_dict, compute_groups=groups)
        self.val_cls_metrics = cls_metrics.clone(prefix="val/")
        self.test_cls_metrics = cls_metrics.clone(prefix="test/")

        if self.post_processing is not None:
            self.post_cls_metrics = cls_metrics.clone(prefix="test/post/")

        self.test_id_entropy = Entropy()

        if self.eval_ood:
            ood_metrics = MetricCollection(
                {
                    "AUROC": BinaryAUROC(),
                    "AUPR": BinaryAveragePrecision(),
                    "FPR95": FPR95(pos_label=1),
                },
                compute_groups=[["AUROC", "AUPR"], ["FPR95"]],
            )
            self.test_ood_metrics = ood_metrics.clone(prefix="ood/")
            self.test_ood_entropy = Entropy()

        if self.eval_shift:
            self.test_shift_metrics = cls_metrics.clone(prefix="shift/")

        # metrics for ensembles only
        if self.is_ensemble:
            ens_metrics = MetricCollection(
                {
                    "Disagreement": Disagreement(),
                    "MI": MutualInformation(),
                    "Entropy": Entropy(),
                }
            )

            self.test_id_ens_metrics = ens_metrics.clone(prefix="test/ens_")

            if self.eval_ood:
                self.test_ood_ens_metrics = ens_metrics.clone(prefix="ood/ens_")

            if self.eval_shift:
                self.test_shift_ens_metrics = ens_metrics.clone(prefix="shift/ens_")

        if self.eval_grouping_loss:
            grouping_loss = MetricCollection({"cls/grouping_loss": GroupingLoss()})
            self.val_grouping_loss = grouping_loss.clone(prefix="val/")
            self.test_grouping_loss = grouping_loss.clone(prefix="test/")

    def _init_mixup(self, mixup_params: dict | None) -> Callable:
        """Setup the optional mixup augmentation based on the :attr:`mixup_params` dict.

        Args:
            mixup_params (dict | None): the detailed parameters of the mixup augmentation. None if
                unused.
        """
        if mixup_params is None:
            mixup_params = {}
        mixup_params = MIXUP_PARAMS | mixup_params
        self.mixup_params = mixup_params

        if mixup_params["mixup_alpha"] < 0 or mixup_params["cutmix_alpha"] < 0:
            raise ValueError(
                "Cutmix alpha and Mixup alpha must be positive."
                f"Got {mixup_params['mixup_alpha']} and {mixup_params['cutmix_alpha']}."
            )

        if mixup_params["mixtype"] == "timm":
            return timm_Mixup(
                mixup_alpha=mixup_params["mixup_alpha"],
                cutmix_alpha=mixup_params["cutmix_alpha"],
                mode=mixup_params["mixmode"],
                num_classes=self.num_classes,
            )
        if mixup_params["mixtype"] == "mixup":
            return Mixup(
                alpha=mixup_params["mixup_alpha"],
                mode=mixup_params["mixmode"],
                num_classes=self.num_classes,
            )
        if mixup_params["mixtype"] == "mixup_io":
            return MixupIO(
                alpha=mixup_params["mixup_alpha"],
                mode=mixup_params["mixmode"],
                num_classes=self.num_classes,
            )
        if mixup_params["mixtype"] == "regmixup":
            return RegMixup(
                alpha=mixup_params["mixup_alpha"],
                mode=mixup_params["mixmode"],
                num_classes=self.num_classes,
            )
        if mixup_params["mixtype"] == "kernel_warping":
            return WarpingMixup(
                alpha=mixup_params["mixup_alpha"],
                mode=mixup_params["mixmode"],
                num_classes=self.num_classes,
                apply_kernel=True,
                tau_max=mixup_params["kernel_tau_max"],
                tau_std=mixup_params["kernel_tau_std"],
            )
        return Identity()

    def _apply_mixup(self, batch: tuple[Tensor, Tensor]) -> tuple[Tensor, Tensor]:
        """Apply the mixup augmentation on a :attr:`batch` of images.

        Args:
            batch (tuple[Tensor, Tensor]): the images and the corresponding targets.

        Returns:
            tuple[Tensor, Tensor]: the images and the corresponding targets transformed with mixup.
        """
        if not self.is_ensemble:
            if self.mixup_params["mixtype"] == "kernel_warping":
                if self.mixup_params["dist_sim"] == "emb":
                    with torch.no_grad():
                        feats = self.model.feats_forward(batch[0]).detach()
                    batch = self.mixup(*batch, feats)
                else:  # self.mixup_params["dist_sim"] == "inp":
                    batch = self.mixup(*batch, batch[0])
            else:
                batch = self.mixup(*batch)
        return batch

    def configure_optimizers(self) -> Optimizer | dict:
        return self.optim_recipe

    def on_train_start(self) -> None:
        """Put the hyperparameters in tensorboard."""
        if self.logger is not None:  # coverage: ignore
            self.logger.log_hyperparams(
                self.hparams,
            )

    def on_validation_start(self) -> None:
        """Prepare the validation step.

        Update the model's wrapper and the batchnorms if needed.
        """
        if self.needs_epoch_update and not self.trainer.sanity_checking:
            self.model.update_wrapper(self.current_epoch)
            if hasattr(self.model, "need_bn_update"):
                self.model.bn_update(self.trainer.train_dataloader, device=self.device)

    def on_test_start(self) -> None:
        """Prepare the test step.

        Setup the post-processing dataset and fit the post-processing method if needed, prepares
        the storage lists for logit plotting and update the batchnorms if needed.
        """
        if self.post_processing is not None:
            calibration_dataset = (
                self.trainer.datamodule.val_dataloader().dataset
                if self.calibration_set == "val"
                else self.trainer.datamodule.test_dataloader()[0].dataset
            )
            with torch.inference_mode(False):
                self.post_processing.fit(calibration_dataset)

        if self.eval_ood and self.log_plots and isinstance(self.logger, Logger):
            self.id_logit_storage = []
            self.ood_logit_storage = []

        if hasattr(self.model, "need_bn_update"):
            self.model.bn_update(self.trainer.train_dataloader, device=self.device)

    def forward(self, inputs: Tensor, save_feats: bool = False) -> Tensor:
        """Forward pass of the inner model.

        Args:
            inputs (Tensor): input tensor.
            save_feats (bool, optional): whether to store the features or
                not. Defaults to ``False``.

        Note:
            The features are stored in the :attr:`self.features` attribute.
        """
        if save_feats:
            self.features = self.model.feats_forward(inputs)
            if hasattr(self.model, "classification_head"):  # coverage: ignore
                logits = self.model.classification_head(self.features)
            else:
                logits = self.model.linear(self.features)
        else:
            self.features = None
            logits = self.model(inputs)
        return logits

    def training_step(self, batch: tuple[Tensor, Tensor]) -> STEP_OUTPUT:
        """Perform a single training step based on the input tensors.

        Args:
            batch (tuple[Tensor, Tensor]): the training data and their corresponding targets

        Returns:
            Tensor: the loss corresponding to this training step.
        """
        batch = self._apply_mixup(batch)
        inputs, target = self.format_batch_fn(batch)

        if self.is_elbo:
            loss = self.loss(inputs, target)
        else:
            logits = self.forward(inputs)
            # BCEWithLogitsLoss expects float target
            if self.binary_cls and isinstance(self.loss, nn.BCEWithLogitsLoss):
                logits = logits.squeeze(-1)
                target = target.float()

            if not self.is_dec:
                loss = self.loss(logits, target)
            else:
                loss = self.loss(logits, target, self.current_epoch)
        if self.needs_step_update:
            self.model.update_wrapper(self.current_epoch)
        self.log("train_loss", loss, prog_bar=True, logger=True)
        return loss

    def validation_step(self, batch: tuple[Tensor, Tensor]) -> None:
        """Perform a single validation step based on the input tensors.

        Compute the prediction of the model and the value of the metrics on the validation batch.

        Args:
            batch (tuple[Tensor, Tensor]): the validation data and their corresponding targets
        """
        inputs, targets = batch
        logits = self.forward(inputs, save_feats=self.eval_grouping_loss)
        logits = rearrange(logits, "(m b) c -> b m c", b=targets.size(0))

        if self.binary_cls:
            probs_per_est = torch.sigmoid(logits).squeeze(-1)
        else:
            probs_per_est = F.softmax(logits, dim=-1)

        probs = probs_per_est.mean(dim=1)
        self.val_cls_metrics.update(probs, targets)

        if self.eval_grouping_loss:
            self.val_grouping_loss.update(probs, targets, self.features)

    def test_step(
        self,
        batch: tuple[Tensor, Tensor],
        batch_idx: int,
        dataloader_idx: int = 0,
    ) -> None:
        """Perform a single test step based on the input tensors.

        Compute the prediction of the model and the value of the metrics on the test batch. Also
        handle OOD and distribution-shifted images.

        Args:
            batch (tuple[Tensor, Tensor]): the test data and their corresponding targets.
            batch_idx (int): the number of the current batch (unused).
            dataloader_idx (int): 0 if in-distribution, 1 if out-of-distribution and 2 if
                distribution-shifted.
        """
        inputs, targets = batch
        logits = self.forward(inputs, save_feats=self.eval_grouping_loss)
        logits = rearrange(logits, "(m b) c -> b m c", b=targets.size(0))
        probs_per_est = torch.sigmoid(logits) if self.binary_cls else F.softmax(logits, dim=-1)
        probs = probs_per_est.mean(dim=1)
        confs = probs.max(-1)[0]

        if self.ood_criterion == "logit":
            ood_scores = -logits.mean(dim=1).max(dim=-1).values
        elif self.ood_criterion == "energy":
            ood_scores = -logits.mean(dim=1).logsumexp(dim=-1)
        elif self.ood_criterion == "entropy":
            ood_scores = torch.special.entr(probs_per_est).sum(dim=-1).mean(dim=1)
        elif self.ood_criterion == "mi":
            mi_metric = MutualInformation(reduction="none")
            ood_scores = mi_metric(probs_per_est)
        elif self.ood_criterion == "vr":
            vr_metric = VariationRatio(reduction="none", probabilistic=False)
            ood_scores = vr_metric(probs_per_est.transpose(0, 1))
        else:
            ood_scores = -confs

        if dataloader_idx == 0:
            # squeeze if binary classification only for binary metrics
            self.test_cls_metrics.update(
                probs.squeeze(-1) if self.binary_cls else probs,
                targets,
            )
            if self.eval_grouping_loss:
                self.test_grouping_loss.update(probs, targets, self.features)

            self.log_dict(self.test_cls_metrics, on_epoch=True, add_dataloader_idx=False)
            self.test_id_entropy(probs)
            self.log(
                "test/cls/Entropy",
                self.test_id_entropy,
                on_epoch=True,
                add_dataloader_idx=False,
            )

            if self.is_ensemble:
                self.test_id_ens_metrics.update(probs_per_est)

            if self.eval_ood:
                self.test_ood_metrics.update(ood_scores, torch.zeros_like(targets))

            if self.id_logit_storage is not None:
                self.id_logit_storage.append(logits.detach().cpu())

            if self.post_processing is not None:
                pp_logits = self.post_processing(inputs)
                if not isinstance(self.post_processing, LaplaceApprox):
                    pp_probs = F.softmax(pp_logits, dim=-1)
                else:
                    pp_probs = pp_logits
                self.post_cls_metrics.update(pp_probs, targets)

        if self.eval_ood and dataloader_idx == 1:
            self.test_ood_metrics.update(ood_scores, torch.ones_like(targets))
            self.test_ood_entropy(probs)
            self.log(
                "ood/Entropy",
                self.test_ood_entropy,
                on_epoch=True,
                add_dataloader_idx=False,
            )
            if self.is_ensemble:
                self.test_ood_ens_metrics.update(probs_per_est)

            if self.ood_logit_storage is not None:
                self.ood_logit_storage.append(logits.detach().cpu())

        if self.eval_shift and dataloader_idx == (2 if self.eval_ood else 1):
            self.test_shift_metrics.update(probs, targets)
            if self.is_ensemble:
                self.test_shift_ens_metrics.update(probs_per_est)

    def on_validation_epoch_end(self) -> None:
        """Compute and log the values of the collected metrics in `validation_step`."""
        res_dict = self.val_cls_metrics.compute()
        self.log_dict(res_dict, logger=True, sync_dist=True)
        self.log(
            "Acc%",
            res_dict["val/cls/Acc"] * 100,
            prog_bar=True,
            logger=False,
            sync_dist=True,
        )
        self.val_cls_metrics.reset()

        if self.eval_grouping_loss:
            self.log_dict(self.val_grouping_loss.compute(), sync_dist=True)
            self.val_grouping_loss.reset()

    def on_test_epoch_end(self) -> None:
        """Compute, log, and plot the values of the collected metrics in `test_step`."""
        # already logged
        result_dict = self.test_cls_metrics.compute()

        # already logged
        result_dict.update({"test/Entropy": self.test_id_entropy.compute()}, sync_dist=True)

        if self.post_processing is not None:
            tmp_metrics = self.post_cls_metrics.compute()
            self.log_dict(tmp_metrics, sync_dist=True)
            result_dict.update(tmp_metrics)

        if self.eval_grouping_loss:
            self.log_dict(
                self.test_grouping_loss.compute(),
                sync_dist=True,
            )

        if self.is_ensemble:
            tmp_metrics = self.test_id_ens_metrics.compute()
            self.log_dict(tmp_metrics, sync_dist=True)
            result_dict.update(tmp_metrics)

        if self.eval_ood:
            tmp_metrics = self.test_ood_metrics.compute()
            self.log_dict(tmp_metrics, sync_dist=True)
            result_dict.update(tmp_metrics)

            # already logged
            result_dict.update({"ood/Entropy": self.test_ood_entropy.compute()})

            if self.is_ensemble:
                tmp_metrics = self.test_ood_ens_metrics.compute()
                self.log_dict(tmp_metrics, sync_dist=True)
                result_dict.update(tmp_metrics)

        if self.eval_shift:
            tmp_metrics = self.test_shift_metrics.compute()
            shift_severity = self.trainer.datamodule.shift_severity
            tmp_metrics["shift/shift_severity"] = shift_severity
            self.log_dict(tmp_metrics, sync_dist=True)
            result_dict.update(tmp_metrics)

            if self.is_ensemble:
                tmp_metrics = self.test_shift_ens_metrics.compute()
                self.log_dict(tmp_metrics, sync_dist=True)
                result_dict.update(tmp_metrics)

        if isinstance(self.logger, Logger) and self.log_plots:
            self.logger.experiment.add_figure(
                "Reliabity diagram", self.test_cls_metrics["cal/ECE"].plot()[0]
            )
            self.logger.experiment.add_figure(
                "Risk-Coverage curve",
                self.test_cls_metrics["sc/AURC"].plot()[0],
            )
            self.logger.experiment.add_figure(
                "Generalized Risk-Coverage curve",
                self.test_cls_metrics["sc/AUGRC"].plot()[0],
            )

            if self.post_processing is not None:
                self.logger.experiment.add_figure(
                    "Reliabity diagram after calibration",
                    self.post_cls_metrics["cal/ECE"].plot()[0],
                )

            # plot histograms of logits and likelihoods
            if self.eval_ood:
                id_logits = torch.cat(self.id_logit_storage, dim=0)
                ood_logits = torch.cat(self.ood_logit_storage, dim=0)

                id_probs = F.softmax(id_logits, dim=-1)
                ood_probs = F.softmax(ood_logits, dim=-1)

                logits_fig = plot_hist(
                    [
                        id_logits.mean(1).max(-1).values,
                        ood_logits.mean(1).max(-1).values,
                    ],
                    20,
                    "Histogram of the logits",
                )[0]
                probs_fig = plot_hist(
                    [
                        id_probs.mean(1).max(-1).values,
                        ood_probs.mean(1).max(-1).values,
                    ],
                    20,
                    "Histogram of the likelihoods",
                )[0]
                self.logger.experiment.add_figure("Logit Histogram", logits_fig)
                self.logger.experiment.add_figure("Likelihood Histogram", probs_fig)

        if self.save_in_csv:
            self.save_results_to_csv(result_dict)

    def save_results_to_csv(self, results: dict[str, float]) -> None:
        """Save the metric results in a csv.

        Args:
            results (dict[str, float]): the dictionary containing all the values of the metrics.
        """
        if self.logger is not None:
            csv_writer(
                Path(self.logger.log_dir) / "results.csv",
                results,
            )


def _classification_routine_checks(
    model: nn.Module,
    num_classes: int,
    is_ensemble: bool,
    ood_criterion: str,
    eval_grouping_loss: bool,
    num_calibration_bins: int,
    mixup_params: dict | None,
    post_processing: PostProcessing | None,
    format_batch_fn: nn.Module | None,
) -> None:
    """Check the domains of the routine's parameters.

    Args:
        model (nn.Module): the model used to make classification predictions.
        num_classes (int): the number of classes in the dataset.
        is_ensemble (bool): whether the model is an ensemble or a single model.
        ood_criterion (str): the criterion for the binary OOD detection task.
        eval_grouping_loss (bool): whether to evaluate the grouping loss.
        num_calibration_bins (int): the number of bins for the evaluation of the calibration.
        mixup_params (dict | None): the dictionary to setup the mixup augmentation.
        post_processing (PostProcessing | None): the post-processing module.
        format_batch_fn (nn.Module | None): the function for formatting the batch for ensembles.
    """
    if ood_criterion not in [
        "msp",
        "logit",
        "energy",
        "entropy",
        "mi",
        "vr",
    ]:
        raise ValueError(
            "The OOD criterion must be one of 'msp', 'logit', 'energy', 'entropy',"
            f" 'mi' or 'vr'. Got {ood_criterion}."
        )

    if not is_ensemble and ood_criterion in ["mi", "vr"]:
        raise ValueError(
            "You cannot use mutual information or variation ratio with a single model."
        )

    if is_ensemble and eval_grouping_loss:
        raise NotImplementedError(
            "Groupng loss for ensembles is not yet implemented. Raise an issue if needed."
        )

    if num_classes < 1:
        raise ValueError(
            f"The number of classes must be a positive integer >= 1. Got {num_classes}."
        )

    if eval_grouping_loss and not hasattr(model, "feats_forward"):
        raise ValueError(
            "Your model must have a `feats_forward` method to compute the grouping loss."
        )

    if eval_grouping_loss and not (
        hasattr(model, "classification_head") or hasattr(model, "linear")
    ):
        raise ValueError(
            "Your model must have a `classification_head` or `linear` "
            "attribute to compute the grouping loss."
        )

    if num_calibration_bins < 2:
        raise ValueError(f"num_calibration_bins must be at least 2, got {num_calibration_bins}.")

    if mixup_params is not None and isinstance(format_batch_fn, RepeatTarget):
        raise ValueError(
            "Mixup is not supported for ensembles at training time. Please set mixup_params to None."
        )

    if post_processing is not None and is_ensemble:
        raise ValueError(
            "Ensembles and post-processing methods cannot be used together. Raise an issue if needed."
        )



import logging
from collections.abc import Callable
from pathlib import Path
from typing import Any

from torchvision.datasets import ImageFolder
from torchvision.datasets.utils import (
    check_integrity,
    download_file_from_google_drive,
    extract_archive,
)


class Fractals(ImageFolder):
    """Dataset used for PixMix augmentations.

    Args:
        root (str): Root directory of dataset.

    Note:
        There is no information on the license of the dataset. It may not
        be suitable for commercial use.
    """

    file_id = "1qC2gIUx9ARU7zhgI4IwGD3YcFhm8J4cA"
    filename = "fractals_and_fvis.tar"
    tgz_md5 = "3619fb7e2c76130749d97913fdd3ab27"

    def __init__(
        self,
        root: str | Path,
        transform: Callable[..., Any] | None = None,
        target_transform: Callable[..., Any] | None = None,
        download: bool = False,
    ) -> None:
        self.root = Path(root)

        if download:
            self.download()

        if not self._check_integrity():
            raise RuntimeError(
                "Dataset not found or corrupted. You can use download=True to download it."
            )

        super().__init__(self.root, transform=transform, target_transform=target_transform)

    def _check_integrity(self) -> bool:
        fpath = self.root / self.filename
        return check_integrity(
            fpath,
            self.tgz_md5,
        )

    def download(self) -> None:
        if self._check_integrity():
            logging.info("Files already downloaded and verified")
            return

        download_file_from_google_drive(
            file_id=self.file_id,
            root=self.root,
            filename=self.filename,
            md5=self.tgz_md5,
        )
        extract_archive(self.root / self.filename, self.root)

    def __getitem__(self, index: int) -> tuple[Any, Any]:
        """Get the samples and targets of the dataset.

        Args:
            index (int): The index of the sample to get.
        """
        return super().__getitem__(index)[0]



from typing import Literal

import torch
from torch import Tensor
from torchmetrics import Metric
from torchmetrics.utilities import rank_zero_warn
from torchmetrics.utilities.data import dim_zero_cat


class VariationRatio(Metric):
    full_state_update = False
    is_differentiable = True
    higher_is_better = False

    def __init__(
        self,
        probabilistic: bool = True,
        reduction: Literal["mean", "sum", "none", None] = "mean",
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        allowed_reduction = ("sum", "mean", "none", None)
        if reduction not in allowed_reduction:
            raise ValueError(
                "Expected argument `reduction` to be one of ",
                f"{allowed_reduction} but got {reduction}",
            )

        self.probabilistic = probabilistic
        self.reduction = reduction

        self.add_state("probs", [], dist_reduce_fx="cat")

        rank_zero_warn(
            "Metric `VariationRatio` will save all predictions in buffer. For "
            " large datasets this may lead to large memory footprint."
        )

    def update(self, probs: Tensor) -> None:
        # store data as (example, estimator, class)
        self.probs.append(probs.transpose(0, 1))

    def compute(self) -> Tensor:
        r"""Computes the variation ratio which amounts to the proportion of
        predicted class labels which are not the chosen class.

        Returns:
            Tensor: Mean disagreement between estimators.
        """
        probs_per_est = dim_zero_cat(self.probs)
        n_estimators = probs_per_est.shape[1]
        probs = probs_per_est.mean(dim=1)

        # best class for example
        max_classes = probs.argmax(dim=-1)

        if self.probabilistic:
            probs_per_est = probs_per_est.permute((0, 2, 1))
            variation_ratio = 1 - probs_per_est[
                torch.arange(probs_per_est.size(0)), max_classes
            ].mean(dim=1)
        else:
            # best class for (example, estimator)
            max_classes_per_est = probs_per_est.argmax(dim=-1)
            variation_ratio = (
                1
                - torch.sum(max_classes_per_est == max_classes.unsqueeze(1), dim=-1) / n_estimators
            )

        if self.reduction == "mean":
            variation_ratio = variation_ratio.mean()
        elif self.reduction == "sum":
            variation_ratio = variation_ratio.sum()
        else:  # None
            pass

        return variation_ratio



import pytest
import torch

from tests._dummies.model import dummy_model
from torch_uncertainty.models import MCDropout, mc_dropout


class TestMCDropout:
    """Testing the MC Dropout class."""

    def test_mc_dropout_train(self):
        model = dummy_model(10, 5, 0.1)
        dropout_model = mc_dropout(model, num_estimators=5)
        dropout_model.train()
        assert dropout_model.training
        dropout_model(torch.rand(1, 10))

        dropout_model = mc_dropout(model, num_estimators=5, last_layer=True)
        dropout_model.train()
        assert dropout_model.training
        dropout_model(torch.rand(1, 10))

    def test_mc_dropout_eval(self):
        model = dummy_model(10, 5, 0.1)
        dropout_model = mc_dropout(model, num_estimators=5)
        dropout_model.eval()
        assert not dropout_model.training
        dropout_model(torch.rand(1, 10))

        dropout_model = mc_dropout(model, num_estimators=5, on_batch=False)
        dropout_model.eval()
        assert not dropout_model.training
        dropout_model(torch.rand(1, 10))

    def test_mc_dropout_errors(self):
        model = dummy_model(10, 5, 0.1)

        with pytest.raises(ValueError, match="`num_estimators` must be strictly positive"):
            MCDropout(model=model, num_estimators=-1, last_layer=True, on_batch=True)

        dropout_model = mc_dropout(model, 5)
        with pytest.raises(TypeError, match="Training mode is expected to be boolean"):
            dropout_model.train(mode=1)

        with pytest.raises(TypeError, match="Training mode is expected to be boolean"):
            dropout_model.train(mode=None)

        model = dummy_model(10, 5, 0.0)
        with pytest.raises(
            ValueError,
            match="At least one dropout module must have a dropout rate",
        ):
            dropout_model = mc_dropout(model, 5)

        model = dummy_model(10, 5, dropout_rate=0)
        del model.dropout
        with pytest.raises(ValueError):
            dropout_model = mc_dropout(model, 5)



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html
from datetime import datetime

import tu_sphinx_theme

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "TorchUncertainty"

copyright = (  # noqa: A001
    f"{datetime.now().year!s}, Adrien Lafage and Olivier Laurent"
)
author = "Adrien Lafage and Olivier Laurent"
release = "0.4.2"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    # "sphinxcontrib.katex",
    "sphinx.ext.mathjax",
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    # "sphinx.ext.intersphinx",  # for links to the API in the tutorials
    "sphinx.ext.napoleon",
    "sphinx.ext.viewcode",
    "sphinx_copybutton",
    "sphinx_codeautolink",
    "sphinx_gallery.gen_gallery",
    # "sphinx_gallery.load_style",
]

sphinx_gallery_conf = {
    "examples_dirs": ["../../auto_tutorials_source"],
    "gallery_dirs": "auto_tutorials",
    "filename_pattern": r"tutorial_",
    "plot_gallery": "True",
    "promote_jupyter_magic": True,
    "backreferences_dir": None,
    "first_notebook_cell": (
        "# For tips on running notebooks in Google Colab, see\n"
        "# https://pytorch.org/tutorials/beginner/colab\n"
        "%matplotlib inline"
    ),
    "reference_url": {
        "sphinx_gallery": None,
    },
}

# Use both the docstrings of the init and the class
autoclass_content = "both"

autosummary_generate = True
napoleon_use_ivar = True

# Disable docstring inheritance
autodoc_inherit_docstrings = False

# Disable displaying type annotations, these can be very verbose
autodoc_typehints = "none"

templates_path = ["_templates"]
exclude_patterns = []

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "tu_sphinx_theme"
html_theme_path = [tu_sphinx_theme.get_html_theme_path()]

html_theme_options = {
    "logo_url": "https://torch-uncertainty.github.io/",
    "menu": [
        {
            "name": "GitHub",
            "url": "https://github.com/ENSTA-U2IS-AI/torch-uncertainty",
        }
    ],
    "pytorch_project": "tutorials",
}

html_static_path = ["_static"]
html_style = "css/custom.css"



import matplotlib.pyplot as plt
import pytest
import torch

from torch_uncertainty.metrics import AdaptiveCalibrationError, CalibrationError


class TestCalibrationError:
    """Testing the CalibrationError metric class."""

    def test_plot_binary(self) -> None:
        metric = CalibrationError(task="binary", num_bins=2, norm="l1")
        metric.update(
            torch.as_tensor([0.25, 0.25, 0.55, 0.75, 0.75]),
            torch.as_tensor([0, 0, 1, 1, 1]),
        )
        fig, ax = metric.plot()
        assert isinstance(fig, plt.Figure)
        assert ax[0].get_xlabel() == "Top-class Confidence (%)"
        assert ax[0].get_ylabel() == "Success Rate (%)"
        assert ax[1].get_xlabel() == "Top-class Confidence (%)"
        assert ax[1].get_ylabel() == "Density (%)"

        plt.close(fig)

    def test_plot_multiclass(
        self,
    ) -> None:
        metric = CalibrationError(task="multiclass", num_bins=3, norm="l1", num_classes=3)
        metric.update(
            torch.as_tensor(
                [
                    [0.25, 0.20, 0.55],
                    [0.55, 0.05, 0.40],
                    [0.10, 0.30, 0.60],
                    [0.90, 0.05, 0.05],
                ]
            ),
            torch.as_tensor([0, 1, 2, 0]),
        )
        fig, ax = metric.plot()
        assert isinstance(fig, plt.Figure)
        assert ax[0].get_xlabel() == "Top-class Confidence (%)"
        assert ax[0].get_ylabel() == "Success Rate (%)"
        assert ax[1].get_xlabel() == "Top-class Confidence (%)"
        assert ax[1].get_ylabel() == "Density (%)"
        plt.close(fig)

    def test_errors(self) -> None:
        with pytest.raises(TypeError, match="is expected to be `int`"):
            CalibrationError(task="multiclass", num_classes=None)
        with pytest.raises(ValueError, match="`n_bins` does not exist, use `num_bins`."):
            CalibrationError(task="multiclass", num_classes=2, n_bins=1)


class TestAdaptiveCalibrationError:
    """Testing the AdaptiveCalibrationError metric class."""

    def test_main(self) -> None:
        ace = AdaptiveCalibrationError(task="binary", num_bins=2, norm="l1", validate_args=True)

        ace = AdaptiveCalibrationError(task="binary", num_bins=2, norm="l1", validate_args=False)
        ece = CalibrationError(task="binary", num_bins=2, norm="l1")
        ace.update(
            torch.as_tensor([0.35, 0.35, 0.75, 0.75]),
            torch.as_tensor([0, 0, 1, 1]),
        )
        ece.update(
            torch.as_tensor([0.35, 0.35, 0.75, 0.75]),
            torch.as_tensor([0, 0, 1, 1]),
        )
        assert ace.compute().item() == ece.compute().item()

        ace.reset()
        ace.update(
            torch.as_tensor([0.3, 0.24, 0.25, 0.2, 0.8]),
            torch.as_tensor([0, 0, 0, 1, 1]),
        )
        assert ace.compute().item() == pytest.approx(
            3 / 5 * (1 - 1 / 3 * (0.7 + 0.76 + 0.75)) + 2 / 5 * (0.8 - 0.5)
        )

        ace = AdaptiveCalibrationError(
            task="multiclass",
            num_classes=2,
            num_bins=2,
            norm="l2",
            validate_args=True,
        )
        ace.update(
            torch.as_tensor([[0.7, 0.3], [0.76, 0.24], [0.75, 0.25], [0.2, 0.8], [0.8, 0.2]]),
            torch.as_tensor([0, 0, 0, 1, 1]),
        )
        assert ace.compute().item() ** 2 == pytest.approx(
            3 / 5 * (1 - 1 / 3 * (0.7 + 0.76 + 0.75)) ** 2 + 2 / 5 * (0.8 - 0.5) ** 2
        )

        ace = AdaptiveCalibrationError(
            task="multiclass",
            num_classes=2,
            num_bins=2,
            norm="max",
            validate_args=False,
        )
        ace.update(
            torch.as_tensor([[0.7, 0.3], [0.76, 0.24], [0.75, 0.25], [0.2, 0.8], [0.8, 0.2]]),
            torch.as_tensor([0, 0, 0, 1, 1]),
        )
        assert ace.compute().item() == pytest.approx(0.8 - 0.5)

        ace = AdaptiveCalibrationError(task="binary", num_bins=3, norm="l2")
        ece = CalibrationError(task="binary", num_bins=3, norm="l2")

        ace.update(
            torch.as_tensor([0.12, 0.26, 0.70, 0.71, 0.91, 0.92]),
            torch.as_tensor([0, 1, 0, 0, 1, 1]),
        )
        ece.update(
            torch.as_tensor([0.12, 0.26, 0.70, 0.71, 0.91, 0.92]),
            torch.as_tensor([0, 1, 0, 0, 1, 1]),
        )
        assert ace.compute().item() > ece.compute().item()

    def test_errors(self) -> None:
        with pytest.raises(TypeError, match="is expected to be `int`"):
            AdaptiveCalibrationError(task="multiclass", num_classes=None)



from collections.abc import Callable

import torch.nn.functional as F
from torch import Tensor, nn

from torch_uncertainty.layers.bayesian import BayesLinear
from torch_uncertainty.layers.distributions import get_dist_linear_layer
from torch_uncertainty.layers.packed import PackedLinear
from torch_uncertainty.models import StochasticModel

__all__ = ["bayesian_mlp", "mlp", "packed_mlp"]


class _MLP(nn.Module):
    def __init__(
        self,
        in_features: int,
        num_outputs: int,
        hidden_dims: list[int],
        layer: type[nn.Module],
        activation: Callable,
        layer_args: dict,
        dropout_rate: float,
        dist_family: str | None,
        dist_args: dict,
    ) -> None:
        """Multi-layer perceptron class.

        Args:
            in_features (int): Number of input features.
            num_outputs (int): Number of output features.
            hidden_dims (list[int]): Number of features for each hidden layer.
            layer (nn.Module): Layer class.
            activation (Callable): Activation function.
            layer_args (Dict): Arguments for the layer class.
            dropout_rate (float): Dropout probability.
            dist_family (str, optional): Distribution family name. ``None`` means point-wise
                prediction. Defaults to ``None``.
            dist_args (Dict, optional): Arguments for the distribution layer class.
        """
        super().__init__()
        self.activation = activation
        self.dropout_rate = dropout_rate
        layers = nn.ModuleList()

        if len(hidden_dims) == 0:
            if layer == PackedLinear:
                layer_args |= {"first": True, "last": True}

            self.final_layer = layer(
                in_features=in_features, out_features=num_outputs, **layer_args
            )
        else:
            if layer == PackedLinear:
                layer_args |= {"first": True, "last": False}

            layers.append(layer(in_features=in_features, out_features=hidden_dims[0], **layer_args))

            if layer == PackedLinear:
                layer_args |= {"first": False}

            for i in range(1, len(hidden_dims)):
                layers.append(
                    layer(in_features=hidden_dims[i - 1], out_features=hidden_dims[i], **layer_args)
                )

            if layer == PackedLinear:
                layer_args |= {"last": True}

            if dist_family is not None:
                dist_layer_class = get_dist_linear_layer(dist_family)
                self.final_layer = dist_layer_class(
                    base_layer=layer,
                    event_dim=num_outputs,
                    in_features=hidden_dims[-1],
                    **layer_args,
                    **dist_args,
                )
            else:
                self.final_layer = layer(
                    in_features=hidden_dims[-1], out_features=num_outputs, **layer_args
                )

        self.layers = layers

    def forward(self, x: Tensor) -> Tensor | dict[str, Tensor]:
        for layer in self.layers:
            x = F.dropout(layer(x), p=self.dropout_rate, training=self.training)
            x = self.activation(x)
        return self.final_layer(x)


def _mlp(
    stochastic: bool,
    in_features: int,
    num_outputs: int,
    hidden_dims: list[int],
    num_samples: int = 16,
    layer: type[nn.Module] = nn.Linear,
    layer_args: dict | None = None,
    activation: Callable = F.relu,
    dropout_rate: float = 0.0,
    dist_family: str | None = None,
    dist_args: dict | None = None,
) -> _MLP | StochasticModel:
    model = _MLP(
        in_features=in_features,
        num_outputs=num_outputs,
        hidden_dims=hidden_dims,
        layer_args=layer_args or {},
        layer=layer,
        activation=activation,
        dropout_rate=dropout_rate,
        dist_family=dist_family,
        dist_args=dist_args or {},
    )
    if stochastic:
        return StochasticModel(model, num_samples)
    return model


def mlp(
    in_features: int,
    num_outputs: int,
    hidden_dims: list[int],
    activation: Callable = F.relu,
    dropout_rate: float = 0.0,
    dist_family: str | None = None,
    dist_args: dict | None = None,
) -> _MLP:
    """Multi-layer perceptron.

    Args:
        in_features (int): Number of input features.
        num_outputs (int): Number of output features.
        hidden_dims (list[int]): Number of features in each hidden layer.
        activation (Callable, optional): Activation function. Defaults to
            F.relu.
        dropout_rate (float, optional): Dropout probability. Defaults to 0.0.
        dist_family (str, optional): Distribution family. Defaults to None.
        dist_args (Dict, optional): Arguments for the distribution layer class. Defaults
            to None.

    Returns:
        _MLP: A Multi-Layer-Perceptron model.
    """
    return _mlp(
        stochastic=False,
        in_features=in_features,
        num_outputs=num_outputs,
        hidden_dims=hidden_dims,
        activation=activation,
        dropout_rate=dropout_rate,
        dist_family=dist_family,
        dist_args=dist_args,
    )


def packed_mlp(
    in_features: int,
    num_outputs: int,
    hidden_dims: list[int],
    num_estimators: int = 4,
    alpha: float = 2,
    gamma: float = 1,
    activation: Callable = F.relu,
    dropout_rate: float = 0.0,
    dist_family: str | None = None,
    dist_args: dict | None = None,
) -> _MLP:
    layer_args = {
        "num_estimators": num_estimators,
        "alpha": alpha,
        "gamma": gamma,
    }
    return _mlp(
        stochastic=False,
        in_features=in_features,
        num_outputs=num_outputs,
        hidden_dims=hidden_dims,
        layer=PackedLinear,
        activation=activation,
        layer_args=layer_args,
        dropout_rate=dropout_rate,
        dist_family=dist_family,
        dist_args=dist_args,
    )


def bayesian_mlp(
    in_features: int,
    num_outputs: int,
    hidden_dims: list[int],
    num_samples: int = 16,
    activation: Callable = F.relu,
    dropout_rate: float = 0.0,
    dist_family: str | None = None,
    dist_args: dict | None = None,
) -> StochasticModel:
    return _mlp(
        stochastic=True,
        num_samples=num_samples,
        in_features=in_features,
        num_outputs=num_outputs,
        hidden_dims=hidden_dims,
        layer=BayesLinear,
        activation=activation,
        dropout_rate=dropout_rate,
        dist_family=dist_family,
        dist_args=dist_args,
    )
